# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Universal tokamak template class for creating machine specific subclasses
"""
from pathlib import Path
from warnings import warn

import numpy as np
from scipy.interpolate import RectBivariateSpline
from scipy.ndimage import gaussian_filter

from .diag import Dsystem
from ..geometry import RegularGrid


class Tokamak(Dsystem):
    """
    Handles magnetic field and divertor

    Attributes
    ----------
    name : str
        Name of tokamak
    params : dict
        Keyword arguments provided to initialization
    divertor_coord : tuple of float
        Contains bounds for divertor area. (x1, x2, y1, y2)

    """
    def __init__(self, shot, **kwargs):
        super().__init__(shot, **kwargs)
        # self.name = self.__class__.__name__
        self.divertor_coord = (0, 0, 0, 0)
        return

    @property
    def mag_name(self):
        name = type(self).__name__
        return '{nm}_Magnetics_{shot}'.format(nm=name, shot=self.shot)

    def download_mag_field(self, shot, tvec=None, source=None):
        """
        Dowloads magnetic field data from network.

        Parameters
        ----------
        shot : int
        tvec : list-like, optional
            specifies timeslices to be loaded
        source : string, optional
            specifies source for loading magnetics data

        Returns
        -------
        dict
            Psi normalized stored in dict with keys (values, time, z, r)
        """
        raise NotImplementedError('Method download_mag_field should be defined in subclass.')

    def load_mag_field(self, loc):
        """
        Loads magnetic field from hdf file saved at loc

        If loc is path to a directory, default data file name from property mag_name is assumed.

        Parameters
        ----------
        loc : str or Path
            location of hdf file containing magnetic field

        Returns
        -------
        dict
            Psi normalized stored in dict with keys (values, time, z, R)
        """
        loc = Path(loc)
        if loc.is_dir():
            loc = Path(loc) / self.mag_name
        mag_field = np.load(str(loc))
        return mag_field

    @staticmethod
    def save_mag_field(loc, mag_field):
        """
        Saves magnetic field into hdf file.

        Expects to have name included to mag_loc if magname is omitted.
        
        Parameters
        ----------
        loc : str or pathlib.Path
            Path to file
        mag_field : dict, optional
           DataArray with magnetic field
        """
        loc = Path(loc)
        try:
            loc.parent.mkdir(parents=True, exist_ok=True)
            np.savez(str(loc), **mag_field)
        except PermissionError as e:
            warn('Magnetic field was not saved: {}'.format(e))
        return

    def fetch_mag_field(self, shot, mag_path, grid=None, tvec=None, save=False, force_dl=False):
        """
        Loads magnetic field from local storage or database and interpolates it if grid is provided.

        Parameters
        ----------
        shot : int
        mag_path : str or Path
        grid : RegularGrid, optional
            Returned values are interpolated to this grid if provided.
        tvec : numpy.ndarray, optional
            Time vector  of requested  magnetic field for download
        save : bool, optional
        force_dl : bool, optional
            forces download from remote database
        """
        mag_path = Path(mag_path)
        try:
            assert force_dl is False
            mag_loc = mag_path / self.mag_name
            magfield = self.load_mag_field(mag_loc)
        except (TypeError, FileNotFoundError, OSError, AssertionError):
            magfield = self.download_mag_field(shot, tvec)
            if save:
                try:
                    mag_path.mkdir(exist_ok=True)
                    self.save_mag_field(mag_path / self.mag_name, magfield)
                except OSError as e:
                    warn('Magnetics data were not saved: {}'.format(e))
        print('Magnetic field loaded.')
        if grid is not None:
            magfield = self.interpolate_mag_field(magfield, grid, tvec)
            print('Magnetic field interpolated.')
        return magfield

    @staticmethod
    def interpolate_mag_field(magfield, grid, tvec=None):
        """
        Interpolates given magnetic field to provided grid using rectangular bivariate spline.

        If tvec is provided, only one nearest time slice of magnetic field is interpolated for each value in tvec.

        Parameters
        ----------
        magfield : dict
        grid : RegularGrid
        tvec : int, float, list, tuple, np.ndarray, optional

        Returns
        -------
        dict
        """
        if tvec is None:
            tidx = np.arange(magfield['time'].size, dtype=np.int)
        else:
            tidx = np.searchsorted(magfield['time'], tvec)
            tidx = np.unique(tidx)
            
        nslices = tidx.size
        mfs = np.zeros((nslices, grid.nz, grid.nr))
        for i in range(nslices):
            flux = magfield['values'][tidx[i]]
            rbs = RectBivariateSpline(magfield['z'], magfield['r'], flux)
            psi_rbs = rbs(grid.z_center, grid.r_center)
            mfs[i] = psi_rbs
        mf_interpolated = dict()
        mf_interpolated['values'] = mfs
        mf_interpolated['time'] = magfield['time'][tidx]
        mf_interpolated['r'] = grid.r_center
        mf_interpolated['z'] = grid.z_center
        return mf_interpolated

    # TODO move to RegularGrid?
    def get_divertor_area(self, grid, boundary=None):
        """
        Returns mask matrix with True for indexes representing pixels inside 
        predefined divertor area. Size and shape is obtained from :attr:`grid`.

        Parameters
        ----------
        grid : RegularGrid
        boundary : numpy.ndarray, optional
            Boundary matrix with same shape as grid

        Returns
        -------
        numpy.ndarray
            Bool mask matrix for determining divertor pixels with shape ny, nx
        """
        xgrid, ygrid = np.meshgrid(grid.r_center, grid.z_center)
        dcoord = self.divertor_coord
        divertor_area = np.logical_and(xgrid > dcoord[0], xgrid < dcoord[1])
        divertor_area *= np.logical_and(ygrid > dcoord[2], ygrid < dcoord[3])
        if boundary is not None:
            divertor_area = divertor_area * boundary
        return divertor_area

    # TODO move to solver or RegularGrid?
    def get_divertor_chnls(self, gmat, grid, boundary=None):
        """
        Finds channels of geometry matrix that have contributions in divertor area.
        Uses dot product of geometry and divertor area matrices.

        Parameters
        ----------
        gmat : csr_matrix
        grid : RegularGrid
        boundary : numpy.ndarray, optional
            Boundary matrix with same shape as grid

        Returns
        -------
        mask : numpy.ndarray of bool
            mask array with True on channels interfering with predefined divertor area
        """
        div_vec = self.get_divertor_area(grid, boundary).flatten()
        gval = gmat.values.reshape(gmat.chnl.size, -1)
        dotp = gval.dot(div_vec)
        mask = dotp.sum(axis=1) > 0
        return mask

    # TODO relocate to algorithms?
    def compute_weight_matrix(self, grid, diw=0.4, smooth_boundary=None, bdr=None):
        """
        Computes weight matrix. Supports different weight for pixels located in divertor area or those that are
        outside vacuum vessel.

        Parameters
        ----------
        grid : RegularGrid
        diw : float, optional
            Specifies divertor pixels' weight. Default value for divertor is 0.4 and standard weight of pixel is 1.
            If standard weight is used, the divertor area will not have any preference.
        smooth_boundary : float, optional
            Specifies weight for pixels outside of vacuum vessel, that is used for smoothing the boundary.
            Vacuum vessel coordinates are defined by bdr parameter. Recommended value is 0.1
        bdr : tuple of numpy.ndarray
            Contains coordinates of vacuum vessel border. Must be provided if smooth_boundary is requested.

        Returns
        -------
        numpy.ndarray dtype float64
            matrix with pixel weights with same shape as grid (#y pixels, #x pixels)
        """
        nx = grid.nr
        ny = grid.nz
        wm = np.ones((ny, nx))
        if diw != 1:
            divertor_area = self.get_divertor_area(grid)
            wm[divertor_area] = diw
        if smooth_boundary:
            bdm = grid.is_inside(*bdr)
            wm = gaussian_filter(wm, sigma=3, mode='nearest')
            wm = np.where(bdm, wm, smooth_boundary)
        return wm



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Library of custom tkinter widgets that have not been specially designed for displaying or processing tomography results.
"""
import sys
import tkinter as tk
from tkinter import ttk

from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg


class ConsoleText(tk.Frame):
    """
    A Tkinter Text widget that provides a scrolling display of console stdout.
    """

    def __init__(self, master=None, cnf=None, **kw):
        """
        Parameters
        ----------
        master
        cnf
        kw

        Notes
        -----
        See the __init__ for Tkinter.Text for most of this stuff.
        """
        tk.Frame.__init__(self, master)
        if cnf is None:
            cnf = {}
        self.varscrl = tk.BooleanVar(value=True)
        try:
            self.varscrl.set(kw['scroll'])
        except KeyError:
            self.varscrl.set(True)
        self.txt = tk.Text(self, cnf, **kw)
        self.std_orig = sys.stdout
        self.write_orig = sys.stdout.write
        self.sbar = ttk.Scrollbar(self, orient=tk.VERTICAL,
                                  command=self.txt.yview)
        self.txt['yscrollcommand'] = self.sbar.set
        self.txt['state'] = 'disabled'
        self.txt.grid(row=0, column=0, sticky='NSEW')
        self.sbar.grid(row=0, column=1, sticky='NSW')
        self.scrollcb = ttk.Checkbutton(self, onvalue=True,
                                        command=self.set_scrl,
                                        variable=self.varscrl)
        self.scrollcb.grid(row=1, column=1, sticky='W')
        self.scrlbl = ttk.Label(self, text='Scroll')
        self.scrlbl.grid(row=1, column=0, sticky='E')
        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, weight=0)
        self.columnconfigure(0, weight=1)
        self.running = False
        self.scroll = False

    def start(self):
        sys.stdout.write = self.write
        self.running = True

    #        self.log_update()

    def stop(self):
        sys.stdout.write = self.write_orig
        self.running = False

    def set_scrl(self):
        self.scroll = self.varscrl.get()

    def write(self, val):
        self.txt['state'] = 'normal'
        self.txt.insert('end', val)
        if self.varscrl.get():
            self.txt.see('end')
        self.txt['state'] = 'disabled'
        self.update()

    def log_update(self):
        self.update()
        # print('log_up', self.running, flush = True)
        if self.running:
            # print('after called')
            self.after(1000, self.log_update)


class ConsoleWindow(tk.Toplevel):
    """
    Console text embedded in Toplevel window.
    """

    def __init__(self, parent=None, **kw):
        tk.Toplevel.__init__(self, parent, **kw)
        if 'title' in kw.keys():
            self.title(kw['title'])
        else:
            self.title('Log window')
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self.console = ConsoleText(self)
        self.console.grid(row=0, column=0, sticky='NSWE')

    def start(self):
        self.console.start()

    def stop(self):
        self.console.stop()


class GraphWindow(tk.Toplevel):
    """
    Basic window prepared for displaing data sequence with use of matplotlib.
    Contains Figure and Axes for plotting, but no method for plotting or 
    atribute with data.
    By default it consists of listbox with data sequence, frame for buttons,
    canvas for plot and slider for navigation.
    
    Parameters
    ----------
    parent : TKInter widget
        Default is None. Used in GUI interface to specify hierarchy.
    
    Keywords
    --------
    vals : array-like
        Contains data from time vector to be shown in listbox and
        under slider.
    
    slices : int
        Number of time steps. Use ether vals or slices. Use if no time 
        vector data are available. Creates vals as range(slices).
    
    slavechbutton : bool
        Switches whether slave check button will be created. If True a button
        is created, that if unchecked will switch off control from master.
    """
    def __init__(self, parent=None, **kw):
        tk.Toplevel.__init__(self, parent)
        self.bind('<Left>', self.go_prev)
        self.bind('<Right>', self.go_next)
        self.bind('s', self.save)
        self.data = []
        self.grid()
        self.columnconfigure(0, weight=0)
        self.columnconfigure(1, weight=1)
        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, weight=0)
        self.slices = 0
        self.vals = []
        self.init_datarange(**kw)

        self.fig = Figure()
        self.ax = self.fig.add_subplot(111)
        self.gcanvas = FigureCanvasTkAgg(self.fig, master=self)
        self.gcanvas.get_tk_widget().grid(row=0, column=1, sticky='NSW')
        self.gcanvas._tkcanvas.grid(row=0, column=1, sticky='NSWE')

        self.lboxmenu = ttk.Frame(self)
        self.lboxmenu.grid(row=0, column=0, sticky='NSWE', rowspan=2)
        self.lboxmenu.rowconfigure(0, weight=0)
        self.lboxmenu.rowconfigure(1, weight=1)
        self.lboxmenu.rowconfigure(2, weight=0)
        self.lboxlbl = ttk.Label(self.lboxmenu, text='Data slices')
        self.lboxlbl.grid(row=0, column=0, sticky='NSWE')
        self.lbvar = tk.StringVar()
        self.lbvar.set([i for i in self.vals])
        self.lbox = tk.Listbox(self.lboxmenu, width=5)
        self.lbox['listvariable'] = self.lbvar
        self.sbar = ttk.Scrollbar(self.lboxmenu, orient=tk.VERTICAL,
                                  command=self.lbox.yview)
        self.lbox['yscrollcommand'] = self.sbar.set
        self.lbox.activate(0)
        self.lbox.bind('<Double-Button-1>', self.lbox_selection)
        self.lbox.bind('<Return>', self.lbox_selection)
        self.lbox.grid(row=1, column=0, sticky='NSWE')
        self.sbar.grid(row=1, column=1, sticky='NSW')
        self.bb_row = 0
        self.bbox = ttk.Frame(self.lboxmenu)
        self.bbox.grid(row=2, column=0, sticky='NSWE')
        self.but = ttk.Button(self.bbox, text='Save',
                              command=self.save)
        self.but.grid(row=self.bb_row, column=0, sticky='N')
        self.bb_row_raise()
        try:
            assert kw['slavechb'] is True
            self.slave = tk.BooleanVar()
            self.slave.set(True)
            slavechb = ttk.Checkbutton(master=self.bbox, text='Slave', variable=self.slave, onvalue=True)
            self.slavechb = slavechb
            self.slavechb.grid(row=self.bb_row, column=0, sticky='W')
            self.bb_row_raise()
        except (KeyError, AssertionError):
            self.slave = tk.BooleanVar(value=True)

        self.slider = ttk.Frame(self)
        self.slider.columnconfigure(1, weight=1)
        self.slider.rowconfigure(0, weight=0)
        self.slider.rowconfigure(1, weight=0)
        self.slider.grid(sticky='NSWE')
        self.scale = tk.Scale(self.slider, orient=tk.HORIZONTAL,
                              showvalue=False,
                              from_=0, to=self.slices - 1,
                              command=self.slider_move,
                              )
        self.label = ttk.Label(self.slider, text='start')
        self.next = ttk.Button(self.slider, text='>', width=2,
                               command=self.go_next)
        self.prev = ttk.Button(self.slider, text='<', width=2,
                               command=self.go_prev)
        self.slider.grid(row=1, column=1)
        self.label.grid(row=1, column=1, sticky='N')
        self.scale.grid(row=0, column=1, sticky='WE')
        self.next.grid(row=0, column=2)
        self.prev.grid(row=0, column=0)
        self.label['text'] = self.vals[0]

    def bb_row_raise(self):
        """
        Increases bb_row, atribute that describes number of placed buttons in
        button box frame widget positioned under listbox. Used for automated 
        row placement.
        Should be called after placing new button to bbox.
        """
        self.bb_row += 1

    def init_datarange(self, **kw):
        """
        Prepares vals or slices for widget construction.
        """
        if 'vals' in kw.keys():
            self.vals = kw['vals']
            self.slices = len(self.vals)
        else:
            self.set_slices(**kw)
            self.vals = [i + 1 for i in range(self.slices)]

    def set_slices(self, **kw):
        """
        Checks keywords for slices infromation. Sets slices to specified number
        or 1 if number of slices was not specified.
        """
        if 'slices' in kw.keys():
            val = kw['slices']
        else:
            val = 1
        self.slices = val

    def set_vals(self, **kw):
        """
        Rutine to set values for listbox menu and displayed numbers and range
        of scale widget.
        """
        if 'vals' in kw.keys():
            self.vals = kw['vals']
            self.slices = len(self.vals)
            self.scale['to'] = self.slices - 1
        else:
            self.set_slices(**kw)
            self.scale['to'] = self.slices - 1
            self.vals = [i + 1 for i in range(self.slices)]
        self.lbvar.set([i for i in self.vals])
        self.label['text'] = self.vals[self.scale.get()]

    def go_to(self, val, *args):
        """
        Puts slider into position given by 'val'.
        
        Parameters
        ----------
        val : int
            Nunber of slider position.
        """
        self.scale.set(val)
        return

    def slider_move(self, *args):
        """
        Executed when position of slider is changed. Changes label of slider,
        plots apropirate data and moves in listbox widget.
        """
        num = self.scale.get()
        self.label['text'] = self.vals[num]
        self.lbox.activate(num)
        self.lbox.see(num)
        self.plot(num)
        return

    def go_next(self, *args):
        new = self.scale.get() + 1
        self.go_to(new)
        return

    def go_prev(self, *args):
        new = self.scale.get() - 1
        self.go_to(new)
        return

    def lbox_selection(self, *args):
        """
        Executed when selection in listbox was executed.
        """
        val = self.lbox.curselection()[0]
        self.go_to(val)
        return

    def save(self, *args):
        raise NotImplementedError()

    def update_graph(self):
        """
        Plots data slice apropirate to slider position.
        """
        val = self.scale.get()
        self.plot(val)

    def update_data(self, ndata, **kw):
        """
        Loads new data.
        
        Parameters
        ----------
        ndata : data structure
            New data to be loaded to Graph window.
        """
        self.data = ndata
        self.set_vals(**kw)

    def plot(self, num):
        """

        Parameters
        ----------
        num

        Returns
        -------

        """
        raise NotImplementedError


class MasterControls(tk.Toplevel):
    def __init__(self, parent, slices, *controled, **kw):
        tk.Toplevel.__init__(self, parent)
        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)
        self.title('Master controls')
        self.grid()
        self.ctrl = controled
        self.slices = slices
        self.vals = []
        self.set_vals(**kw)
        self.slider = ttk.Frame(self)
        self.slider.columnconfigure(1, weight=1)
        self.slider.rowconfigure(0, weight=1)
        self.slider.rowconfigure(1, weight=0)
        self.slider.rowconfigure(2, weight=1)
        self.slider.grid(row=0, column=0, sticky='NSWE')
        self.scale = tk.Scale(self.slider, orient=tk.HORIZONTAL,
                              showvalue=False,
                              from_=0, to=slices - 1,
                              command=self.slider_move)
        self.scale.grid(row=1, column=1, sticky='WE')
        self.label = ttk.Label(self.slider, text=self.vals[0])
        self.label.grid(row=2, column=1, sticky='N')
        self.next = ttk.Button(self.slider, text='->',
                               command=self.go_next)
        self.next.grid(row=1, column=2)
        self.prev = ttk.Button(self.slider, text='<-',
                               command=self.go_prev)
        self.prev.grid(row=1, column=0)
        self.syncvar = tk.BooleanVar()
        self.syncvar.set(1)
        sync = ttk.Checkbutton(self.slider, text='Synchronize',
                               #                                   onvalue = True,
                               variable=self.syncvar)
        sync.grid(row=0, columnspan=3, sticky='S')
        self.bind('<Left>', self.go_prev)
        self.bind('<Right>', self.go_next)
        self.bind('<Return>', self.change_sync)
        self.nogo = ttk.Label(self, text='Create graphs to be controled '
                                         'before creating master controls')
        if len(controled) == 0:
            self.nogo.grid(sticky='NSWE')
        else:
            self.nogo.destroy()

    def set_vals(self, **kw):
        if 'vals' in kw.keys():
            self.vals = kw['vals']
        else:
            self.vals = [i + 1 for i in range(self.slices)]

    # def set_ctrl(self, control):
    #     self.ctrl = control

    def set_slices(self, slices, **kw):
        self.slices = slices
        self.set_vals(**kw)
        self.scale['to'] = slices - 1

    def change_sync(self, *args):
        state = self.syncvar.get()
        self.syncvar.set(not state)

    def slider_move(self, *args):
        num = self.scale.get()
        self.label['text'] = self.vals[num]
        for i in self.ctrl:
            try:
                if i.slave.get():
                    i.go_to(num)
            except AttributeError:
                self.ctrl = self.master.get_ctrl()

    def go_next(self, *args):
        if self.syncvar.get():
            new = self.scale.get() + 1
            self.scale.set(new)
        else:
            for i in self.ctrl:
                if i.slave.get():
                    i.go_next()

    def go_prev(self, *args):
        if self.syncvar.get():
            new = self.scale.get() - 1
            self.scale.set(new)
        else:
            for i in self.ctrl:
                if i.slave.get():
                    i.go_prev()



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Implementation of Phillips-Tikhonov regularisation scheme
"""

import time

import numpy as np
import scipy.sparse as sparse
from scipy.sparse.linalg import spsolve
from scipy.optimize import minimize_scalar


class Tikhonov(object):
    r"""
    Template class that implements Phillips-Tikhonov regularisation scheme.
    
    Does not implement any method for obtaining regularisation matrix nor inversion run

    Attributes
    ----------
    last_chi : float
        person test result of the last computed inversion
    logalpha : float
        :math:`\log(\alpha)` logarithm of regularisation parameter
    gmat : numpy.ndarray
        geometry matrix with shape (#channels, #nodes)
    signal : numpy.ndarray
    gdg : numpy.ndarray
       :math:`\mathbf{T}^T \cdot \mathbf{T}` symmetrised geometry matrix, shape (#nodes, #nodes)
    gdsig : numpy.ndarray
        right side of regularisation scheme :math:`\mathbf{T}^T \cdot \mathbf{f}`
    """
    def __init__(self):
        super().__init__()
        self.last_chi = None
        self.logalpha = None
        self.gmat = None
        self.signal = None
        self.gdg = None
        self.gdsig = None

    @staticmethod
    def solve(a, b):
        r"""
        Finds solution of :math:`\mathbf{Ax}=\mathbf{b}` using scipy.sparse.linalg.spsolve
        """
        return spsolve(a, b)

    def invert(self, signals, gmat, derivatives, w_factor=None, mfi_num=3, bounds=(-15, 0), iter_max=10, w_max=1,
               aniso=0):
        """
        Inverses normalised signals using `mfi_num` Fisher Information cycles each with `iter_max` steps of regularisation parameter optimisation.

        See smoothing_mat documentation for more information about derivative matrix formats.

        Parameters
        ----------
        signals : numpy.ndarray
            error normalised signals on detector channels
        gmat : scipy.sparse.csr_matrix
            geometry matrix normalised by estimated errors
        derivatives : list
            list of tuples containing pairs of sparse derivatives matrices
        w_factor : numpy.ndarray, optional
            weight matrix multipliers with shape (#nodes, )
        mfi_num : float, optional
            number of MFR iterations
        bounds : tuple of two floats, optional
            exponent values for bounds of regularisation parameter alpha
        iter_max : float, optional
            maximum number of root finding iterations
        w_max : float, optional
            value used in weigh matrix for zero or negative nodes
        aniso : float
            Determines anisotropy of derivative matrices, passed to sigmoid function

        Returns
        -------
        numpy.ndarray
            vector with nodes emissivities, shape (#nodes, )
        

        See Also
        --------
        smoothing_mat
        """
        ela = time.time()
        self.signal = signals
        self.gmat = gmat
        self.gdg = gmat.T.dot(gmat)
        self.gdsig = gmat.T.dot(signals)
        npix = gmat.shape[1]
        g = np.ones(npix)
        mfi_count = 0
        while mfi_count < mfi_num:
            w = 1 / g
            w[w < 0] = w_max
            w = sparse.diags(w)
            if w_factor is not None:
                w = w * sparse.diags(w_factor)
            objective = self.smoothing_mat(w, derivatives, aniso)
            res = minimize_scalar(self.test_regularization,
                                  method='bounded',
                                  bounds=bounds,
                                  args=objective,
                                  options={'maxiter': iter_max},
                                  )
            self.logalpha = res.x
            m = self.gdg + 10 ** res.x * objective
            g = self.solve(m, self.gdsig)
            mfi_count += 1
        ela = time.time() - ela
        print('last chi^2 = {:.4f}, time: {:.2f} s'.format(self.last_chi, ela))
        return g

    def smoothing_mat(self, w, derivatives, aniso):
        """
        Computes smoothing matrix from provided derivative matrices and weight factors determined by emissivity from
        previous iteration of minimisation of Fisher Information.

        Multiple pairs of derivatives matrices computed by different numerical scheme can be used. Each pair should
        contain derivatives in each locally orthogonal direction.

        Parameters
        ----------
        w : scipy.sparse.dia.dia_matrix
            (#nodes, #nodes) diagonal matrix with pixel weight factors
        derivatives : list of scipy.sparse.csc_matrix pairs
            contains sparse derivative matrices, each pair contains derivatives in both locally orthogonal coordinates
        aniso : float
            anistropic factor, positive values make derivatives along first coordinate more significant

        Returns
        -------
        smooth
            smoothing matrix
        """
        s1 = 0
        s2 = 0
        for pair in derivatives:
            s1 = s1 + pair[0].T.dot(w).dot(pair[0])
            s2 = s2 + pair[1].T.dot(w).dot(pair[1])
        smooth = self.sigmoid(aniso) * s1 + self.sigmoid(-aniso) * s2
        return smooth

    @staticmethod
    def sigmoid(x):
        return 1 / (1 + np.exp(-x))

    def pearson_test(self, g):
        r"""
        Computes retrofit and residuum :math:`\chi^2` using pearson test

        .. math ::
            \chi^2 = \frac{1}{M} \sum_{i}^{M} \left(\tilde{\mathbf{f}} - \tilde{\mathbf{T}} \cdot \mathbf{g} \right)_i^2

        Parameters
        ----------
        g : numpy.ndarray
            vector of tested emissivity

        Returns
        -------
        numpy.float64
        """
        retrofit = self.gmat.dot(g)
        misfit = retrofit - self.signal
        misfit_sq = np.power(misfit, 2)
        res = np.average(misfit_sq)
        return res

    def __call__(self, data, gmat, dmats, errors, **kwargs):
        """
        Normalises signal and geometry matrix using estimated errors of measurement and
        then executes a sequence of MFR reconstructions.

        Parameters
        ----------
        data : numpy.ndarray
            input to be inverted with shape (#time slices, #channels)
        gmat : sparse.csr_matrix
            shape (#channels, #nodes)
        dmats : list
            list of tuples with derivative matrix pairs
        errors : int, float or np.ndarray
            Can have shapes (#channels, ), (#time slices,) or (#tslices, #channels)
        
        Returns
        -------
        res : np.ndarray
            tomographic reconstruction, with shape (#time slices, #nodes)
        chi : np.ndarray
            Pearson test values for final results, shape (#time slices, )
        """
        if np.ndim(data) == 1:  # flat data, assume one time slice
            data = data.reshape(1, -1)
        nslices = data.shape[0]
        nchnls = data.shape[1]
        nnodes = gmat.shape[1]

        if np.ndim(errors) == 1:
            if errors.size == nslices:
                errors = errors.reshape(-1, 1)
            elif errors.size == nchnls:
                errors = errors.reshape(1, -1)
            else:
                raise ValueError('Incompatible shape of errors {} with the data {}'.format(errors.shape, data.shape))
        
        errors = errors * np.ones_like(data)
        signal_nrm = data / errors

        res = np.empty((nslices, nnodes))
        chi = np.empty(nslices)

        for i in range(nslices):
            signal_np = signal_nrm[i, :]
            error_sp = sparse.diags(1/errors[i, :])
            gmat_nrm = error_sp.dot(gmat)
            res[i] = self.invert(signal_np, gmat_nrm, dmats, **kwargs)
            chi[i] = self.last_chi

        return res, chi



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
from os import path
from .derivative import compute_aniso_dmats, compute_iso_dmats
from .geometry import *
from .inversions import *
from .io import *
from .phantoms import *
from .tools import *

with open(path.join(path.dirname(__file__), 'VERSION')) as version_file:
    __version__ = version_file.read().strip()



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Contains various tools to ease using the package.

 - DataHandler class prototype
 - projection space plot of a system
"""
from .projection import projection_space
from .data import DataHandler



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Averaging of emissivity in bins shaped according to magnetic flux reconstruction
"""
import os
import tkinter as tk
from tkinter import ttk

import numpy as np

from tomotok.core.phantoms import iso_psi
from .widgets_tomo import GraphRgb


class GraphProfiler(GraphRgb):
    """
    Based on GraphRgb class. Usefull for observing power profile in respect to
    magentic flux surfaces.
    
    Parameters
    ----------
    parent : TKInter class parent
        Default is None.
    data : dict
        Contains data, default is empty dict.
    
    Keywords
    --------
    title : str
        Name of window.
    """

    def __init__(self, parent=None, **kw):
        #        tk.Toplevel.__init__(self, parent)
        GraphRgb.__init__(self, parent, **kw)
        if not 'title' in kw:
            self.title('Profiler')
        if not hasattr(self, 'mdata'):
            print('Found no magnetics data, using isotropic flux.')
            self.mdata = iso_psi(self.shape[2], self.shape[1])
            self.mdata = self.mdata.reshape(1, self.shape[1], self.shape[2])
            self.mtvec = np.zeros(1)
        self.ax_orig = self.ax.get_position()
        self.ax2 = self.fig.add_subplot(122)
        self.ax2.set_axis_off()
        dummy_ax = self.fig.add_subplot(121)
        self.ax_left = dummy_ax.get_position()
        dummy_ax.remove()

        self.pprop = ttk.Frame(self, padding=(3, 3, 3, 3))  # , relief = 'solid')
        xrow = 1
        self.pprop.grid(row=0, column=2, rowspan=2, sticky='NSWE')
        self.stavar = tk.StringVar(value=0)
        self.stasbox = tk.Spinbox(self.pprop, from_=0, to=1,
                                  increment=.1, textvariable=self.stavar,
                                  command=self.plot_contour, width=5)
        self.stasbox.grid(row=xrow, column=1)
        self.stalbl = ttk.Label(self.pprop, text='start')
        self.stalbl.grid(row=xrow, column=0)

        self.stovar = tk.StringVar(value=1)
        self.stosbox = tk.Spinbox(self.pprop, from_=0, to=1.5,
                                  increment=.1, textvariable=self.stovar,
                                  command=self.plot_contour, width=5)
        self.stosbox.grid(row=xrow + 1, column=1)
        self.stolbl = ttk.Label(self.pprop, text='stop')
        self.stolbl.grid(row=xrow + 1, column=0)

        self.numvar = tk.StringVar(value=10)
        self.numsbox = tk.Spinbox(self.pprop, from_=1, to=50,
                                  increment=1, textvariable=self.numvar,
                                  command=self.plot_contour, width=5)
        self.numsbox.grid(row=xrow + 2, column=1)
        self.numlbl = ttk.Label(self.pprop, text='#')
        self.numlbl.grid(row=xrow + 2, column=0)
        self.cmpbtn = ttk.Button(self.pprop, text='Compute',
                                 command=self.compute)
        self.cmpbtn.grid(row=xrow + 3, column=0, columnspan=2, sticky='WE')
        #        self.levels = self.get_levels()

        self.snvar = tk.StringVar(value=1)
        self.snsbox = tk.Spinbox(self.pprop, from_=1,
                                 to=min(100, len(self.vals)),
                                 increment=1,
                                 textvariable=self.snvar,
                                 width=5)
        self.snsbox.grid(row=xrow + 6, column=1)
        self.snlbl = ttk.Label(self.pprop, text='n')
        self.snlbl.grid(row=xrow + 6, column=0)
        self.snbtn = ttk.Button(self.pprop, text='Save n-th',
                                command=self.save_nth)
        self.snbtn.grid(row=xrow + 7, column=0, columnspan=2)

        self.expanded = False
        self.steps_plt = {}
        self.style = {'bol': 'solid', 'ntr': 'dashdot', 'sxr': 'dashed'}

    def get_levels(self, *args):
        """
        Gets levels from widgets for plotting contours
        """
        sta = np.float(self.stasbox.get())
        sto = np.float(self.stosbox.get())
        num = np.int(self.numsbox.get())
        return np.linspace(sta, sto, num)

    def plot_contour(self, *args):
        """
        Plots contours to highlight flux locations separating profiled regions.
        """
        self.ax.collections.clear()
        self.levels = self.get_levels()
        val = self.scale.get()
        mval = self.get_mval(val)
        self.cont = self.ax.contour(self.mdata[mval],
                                    levels=self.levels,
                                    extent=self.extent,
                                    # colors='w',
                                    colors='k',
                                    linewidths=1,
                                    )
        if self.magvar.get():
            mval = self.get_mval(val)
            self.mcont = self.ax.contour(self.mdata[mval], [1.0],
                                         extent=self.extent,
                                         # colors='w',
                                         colors='k',
                                         linewidths=1,
                                         linestyles='dashed')
        self.gcanvas.draw()

    #    def compute_cmd(self, *args):
    #        self.compute(*args)
    #        self.gcanvas.draw()

    def compute(self, *args):
        """
        Computes radiated power in given regions.
        """
        self.ax.set_position(self.ax_left)
        self.ax2.set_axis_on()
        self.ax2.set_xlabel('Psi normalized [-]')
        #        self.ax2.set_ylabel('Pixel normalized power [-]')
        self.ax2.set_ylabel('Normalized power density [-]')
        self.ax2.yaxis.tick_right()
        self.ax2.yaxis.set_label_position('right')

        self.res = []
        tmx = 0
        for i, val in enumerate(self.vals):
            mx = 0
            res = {}
            mval = self.get_mval(i)
            for di in self.diags:
                res[di] = []
            rgbind = list(self.get_rgbval(i))
            tind = dict(zip(['bol', 'ntr', 'sxr'], rgbind))
            for l in range(len(self.levels) - 1):
                ge = self.mdata[mval] >= self.levels[l]
                le = self.mdata[mval] < self.levels[l + 1]
                inds = le * ge
                for d in self.diags:
                    #                res[d].append(np.sum(self.data[d][inds,tind[d]]))
                    res[d].append(np.average(self.data[d][tind[d], inds]))
                    mx = np.max([mx, np.max(res[d])])
            self.res.append(res)
            tmx = max(tmx, mx)
        self.centers = np.diff(self.levels)[0] / 2 + self.levels[:-1]

        self.ax.legend()

        self.y2max = tmx
        self.ax2.set_ylim(0, 1.05 * self.y2max)
        self.ax2.set_xlim(float(self.stasbox.get()), float(self.stosbox.get()))
        #        if not self.expanded:
        #            self.gcanvas.draw()
        self.expanded = True
        self.plot(self.scale.get())
        self.ax2.legend()

    def save(self, fname='Graphs/'):
        if not os.path.exists(fname):
            os.makedirs(fname)
        v = self.scale.get()
        sta = self.stasbox.get()
        sto = self.stosbox.get()
        num = self.numsbox.get()
        self.fig.savefig(fname +
                         '{} at {:.0f}ms {}-{}-{}psi profiler.png'.format(
                         # '{} at {:.0f}ms {}-{}-{}psi profiler.eps'.format(
                             self.shot, 1000 * self.vals[v], sta, sto, num))

    def update_data(self, ndata, **kw):
        GraphRgb.update_data(self, ndata, **kw)
        self.ax.lines.clear()
        for i in self.diags:
            self.ax.plot([], c=self.col[i], ls='', marker='s', label=i)
        self.ax.legend()
        self.gcanvas.draw()

    def plot(self, val, *args):
        inds = self.get_rgbval(val)
        b = self.data['bol'][inds[0]]
        n = self.data['ntr'][inds[1]]
        s = self.data['sxr'][inds[2]]
        rgb = np.stack((b, n, s), axis=2)
        rgb = rgb / np.max(rgb)
        cmy = np.ones_like(rgb) - rgb
        rgb = cmy
        if self.grids[0]:
            self.ax.set_xlabel('R [m]')
        else:
            self.ax.set_xlabel('Pixel')
        if self.grids[0]:
            self.ax.set_ylabel('z [m]')
        else:
            self.ax.set_ylabel('Pixel')

        if len(self.ax.images) == 0:
            #            self.ax.imshow(cmy, origin = 'lower',
            self.ax.imshow(rgb, origin='lower',
                           extent=self.extent
                           )
        #            shp = np.shape(rgb)
        #            self.ax.set_xlim(0,shp[1])
        #            self.ax.set_ylim(0,shp[0])
        else:
            self.ax.images[0].set_data(rgb)
            #            self.ax.images[0].set_data(cmy)
            self.ax.images[0].set_extent(self.extent)
        self.plot_contour()
        if self.expanded:
            self.fig.suptitle('#{} Profile @ {:.0f}ms'.format(
                self.shot, 1000 * self.vals[self.scale.get()]))
            self.ax2.lines.clear()
            #            self.ax2.clear()
            #            self.ax2.set_axis_on()
            #            self.ax2.set_xlabel('Psi normalized [-]')
            ##            self.ax2.set_ylabel('Pixel normalized power [-]')
            #            self.ax2.set_ylabel('Normalized power denplot_contoursity [-]')
            #            self.ax2.yaxis.tick_right()
            #            self.ax2.yaxis.set_label_position('right')
            for d in self.diags:
                #                if len(self.ax2.lines) > 0:
                #                    self.ax2.lines.remove(self.steps_plt[d])
                #                self.ax2.plot(self.centers, res[d], c=self.col[d],
                #                          ls=self.style[d], label=d)
                self.steps_plt[d] = self.ax2.step(self.centers,
                                                  self.res[val][d], c=self.col[d],
                                                  where='mid',
                                                  #                                      linestyle=self.style[d],
                                                  label=d)
        self.gcanvas.draw()

    def save_nth(self):
        """
        Saves every nth slice, where n is defined by snvar tk.Stringvar
        asociated with snsbox tk.Spinbox widget.
        """
        i = 0
        n = int(self.snvar.get())
        while i < len(self.vals):
            self.go_to(i)
            self.plot(i)
            self.save(fname='Graphs/prfseq/{}/'.format(self.shot))
            i += n



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Contains functions and classes required for geometry matrix computation using single line of sight approximation.

 - Functions computing contributions of nodes to gmat
 - reconstruction grid definition
 - line of sight generators for artificial diagnostics
 - gmat handler prototype
"""
from .generators import sparse_line_3d, calcam_sparse_line_3d
from .grids import RegularGrid
from .los import generate_los, generate_directions
from .io import save_sparse_gmat, load_sparse_gmat



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Saving and loading functions for geometry matrices in sparse format using h5py module.
"""
from typing import Tuple
import h5py

import scipy.sparse as sparse

from .grids import RegularGrid


def save_sparse_gmat(floc, gmat: sparse.csr_matrix, grid: RegularGrid, attrs_dct: dict=None) -> None:
    """
    Saves geometry matrix together with description of the grid.

    Currently suports only sparse CSR matrices and regular rectangular grids.

    Parameters
    ----------
    floc : str
        file location
    gmat : csr_matrix
        with shape (#chanels, #nodes)
    grid : RegularGrid
        class describing regular grid the gmat was computed for
    attrs_dict : dict, optional
        additional attributes to be saved in hdf file
    """
    if not isinstance(gmat, sparse.csr_matrix):
        msg = 'This function can only save sparse gmat. ' + \
            'For saving matrix in dense format use core.io.to_hdf function'
        raise ValueError(msg)
    with h5py.File(floc, 'w') as f:
        f.attrs['version'] = '0.1'
        f.create_dataset('indices', data=gmat.indices)
        f.create_dataset('indptr', data=gmat.indptr)
        f.create_dataset('data', data=gmat.data)
        f.attrs['format'] = gmat.format
        f.attrs['shape'] = gmat.shape
        grp = f.create_group('grid')
        grp.create_dataset('nr', data=grid.nr)
        grp.create_dataset('nz', data=grid.nz)
        grp.create_dataset('rlims', data=grid.rlims)
        grp.create_dataset('zlims', data=grid.zlims)
        grp.attrs['type'] = 'regular_rectangles'
        if attrs_dct is not None:
            for key in attrs_dct:
                f.attrs[key] = attrs_dct[key]


def load_sparse_gmat(floc: str) -> Tuple[sparse.csr_matrix, RegularGrid]:
    """
    Loads hdf file and creates sparse csr_matrix and RegularGrid class

    Parameters
    ----------
    floc : str
        hdf file location
    
    Returns
    -------
    gmat : sparse.csr_matrix
    grid : geometry.RegularGrid
    """
    with h5py.File(floc, 'r') as fl:
        data = fl['data'][:]
        indptr = fl['indptr'][:]
        indices = fl['indices'][:]
        shp = fl.attrs['shape']
        nx = fl['grid/nr'][()]
        ny = fl['grid/nz'][()]
        xlims = fl['grid/rlims'][()]
        ylims = fl['grid/zlims'][()]
    gmat = sparse.csr_matrix((data, indices, indptr), shape=shp)
    grid = RegularGrid(nx, ny, xlims, ylims)
    return gmat, grid



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Simple hdf saving and loading functions
"""
from warnings import warn

import h5py


def to_hdf(dct, floc, attrs={}):
    """
    Saves provided dict into a hdf file

    Parameters
    ----------
    dct : dict
        [description]
    floc : str
        path to file with name
    attrs : dict
        metadata to be saved to attributes
    """
    warn('Support for dense geometry matrices will be removed in the future.', FutureWarning)
    with h5py.File(floc, 'w') as fl:
        for key in dct:
            fl.create_dataset(str(key), data=dct[key])
        # if name is not None:
        for key in attrs:
            fl.attrs[key] = attrs[key]
    return


def from_hdf(floc):
    dct = {}
    with h5py.File(floc, 'r') as fl:
        for key in fl.keys():
            dct[key] = fl[key][:]
    return dct



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Tool for easier handling of geometry matrices. Work in progress.
"""
import h5py
import warnings
from pathlib import Path


class GmatHandler(object):
    """[summary]

    Attributes
    ----------
    diag
    grid
    gmat : dict
    """

    def __init__(self, diag, grid, gmat=None):
        super().__init__()
        self.diag = diag
        self.grid = grid
        self.gmat = gmat
        rmin = grid.rmin
        rmax = grid.rmax
        zmin = grid.zmin
        zmax = grid.zmax
        nr = grid.nr
        nz = grid.nz
        self.name = 'Gmat_h5_{z}x{r}-{z1}-{z2}-{r1}-{r2}_{gt}'.format(
            r=nr, z=nz, r1=rmin, r2=rmax, z1=zmin, z2=zmax, gt='{}')

    def compute_sparse_lines(self, **kw):
        """
        Computes geometry matrix using single line of sight approx. with numerical algorithm.
        """
        xchords, ychords = self.diag.get_chord_geometry()
        from .generators import sparse_line_3d
        self.gmat = sparse_line_3d(rchord=xchords, vchord=ychords, grid=self.grid, **kw)

    def save_dense(self, gmat_path, grid=None, filename=None):
        """
        Saves geometry matrix to gmat_path. Either grid or filename must be specified.

        If filename is not specified, the default filename is generated using grid properties.

        Parameters
        ----------
        gmat_path : str or Path
            saving folder
        grid : Pixgrid, optional
        filename : str, optional
        """
        gmat_path = Path(gmat_path)
        try:
            assert grid is not None or filename is not None
        except AssertionError:
            raise ValueError('Either grid or filename must be specified.')
        if filename is None:
            gpath = Path(gmat_path) / self.name.format(self.gmat['type'])
        else:
            gpath = Path(gmat_path) / filename
        try:
            gpath.parent.mkdir(parents=True, exist_ok=True)
            with h5py.File(str(gpath), 'w') as hdfgmat:
                for key in ['values', 'z', 'r']:
                    hdfgmat[key] = self.gmat[key]
                hdfgmat.attrs['type'] = self.gmat['type']
        except OSError as e:
            warnings.warn(
                'Can not create or write to gmat_path. Geometry matrix was not saved. {}'.format(e))

    def load_dense(self, gpath, filename=None):
        """
        Loads geometry matrix.

        Assumes that gpath is full path to gmat file if filename is omitted.

        Parameters
        ----------
        gpath : str or Path
        filename : str, optional
            name of file containing geometry matrix

        Returns
        -------
        dict
        """
        gpath = Path(gpath)
        if self.gmat is None:
            self.gmat = {}
        if filename is not None:
            gpath = gpath / filename
        with h5py.File(str(gpath), 'r') as hdfgmat:
            for key in ['values', 'z', 'r']:
                self.gmat[key] = hdfgmat[key][...]
            self.gmat['type'] = hdfgmat.attrs['type']

    def fetch(self, gmat_path='', gtype='line', save=False, force_comp=False):
        """
        Tries to load geometry matrix for specified grid from gmat path.
        Computes new matrix if appropriate matrix is not available.

        Parameters
        ----------
        gmat_path : str or Path, optional
        gtype : str, optional
            Specifies type of geometry matrix to be used. {'line', 'poly'} are supported for computation.
        save : bool, optional
        force_comp : bool, optional
            Forces computation of geometry matrix
        """
        grid = self.grid
        name = self.name.format(gtype)
        gpath = Path(gmat_path) / name
        try:
            assert force_comp is False
            self.load(gpath)
        except (OSError, AssertionError):
            print('Unable to load geom. matrix from {}.'.format(gpath),
                  'Generating new geometry matrix.')
            if gtype == 'line':
                self.compute_lines()
            elif gtype == 'poly':
                self.compute_polygons()
            if save:
                self.save(gmat_path, grid)



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Routines for generation of line of sight start and end points.

Creates the lines of sights in x axis direction and then rotates them about the other horizontal axis, then about vertical axis.
"""
import numpy as np


def generate_directions(num=(10, 1), fov=(45, 0), axis=(1, 0, 0), elong=1.):
    """
    Creates direction vectors for lines of sight using camera like convention.

    The first line of sight is top left, then the numbering follows a row, the last one is bottom right.

    Parameters
    ----------
    num : int or (int,int), optional
        number of generated chords (vertical, horizontal)
    fov : float, or tuple of two float, optional
        vertical or (vertical, horizontal) field of view in degrees
    axis : tuple of three floats, optional
        direction of symmetry axis
    elong : float, optional
        vector length multiplier

    Returns
    -------
    dirs : numpy.ndarray
        contains direction vectors cartesian coordinates, shape (#los, 3)
    """
    if isinstance(fov, (int, float)):
        fov = (fov, 0)
    fov = np.deg2rad(fov)
    try:
        ntot = num[0] * num[1]
    except TypeError:
        ntot = num
        num = (num, 1)

    dirs = np.full((ntot, 3), elong, dtype=np.double)

    ve = elong * np.tan(fov[0] / 2)
    ye = elong * np.tan(fov[1] / 2)
    vg = np.linspace(ve, -ve, num[0])
    yg = np.linspace(-ye, ye, num[1])

    ym, vm = np.meshgrid(yg, vg)
    dirs[:, 1] = ym.flatten()
    dirs[:, 2] = vm.flatten()

    nrm = np.linalg.norm(axis)
    try:
        axis = axis / nrm
    except ZeroDivisionError:
        raise ValueError('Axis has to be a non zero vector.')
    vo = np.arcsin(axis[2])  # vertical offset angle
    if axis[1] == axis[0] == 0:  # vertical vector
        dh = 0
    elif axis[0] == 0:  # horizontal in y-axis direction
        dh = axis[1] * np.inf
    else:  # general direction
        dh = np.sign(axis[1]) * np.abs(axis[1] / axis[0])
    ho = np.arctan(dh)
    dirs = rot_v(dirs, vo)
    dirs = rot_h(dirs, ho)
    if axis[0] < 0:
        dirs[:, 0] = -dirs[:, 0]
    return dirs


def generate_los(pinhole=(0, 0, 0), num=(10, 1), fov=(45, 0), axis=(1, 0, 0), elong=1.):
    """
    Creates line of sight endpoints with uniform distribution.

    Parameters
    ----------
    num : int or (int,int), optional
        number of generated chords (vertical, horizontal)
    pinhole : tuple of three floats, optional
        r and z coordinates of pinhole
    fov : float, or tuple of two float, optional
        vertical and horizontal field of view in degrees
    axis : tuple of three floats, optional
        direction of chordal axis
    elong : float, optional
        chord length multiplier

    Returns
    -------
    start : numpy.ndarray
        array with line of sight start points coordinates, shape (#los, 3)
    end : numpy.ndarray
        array with line of sight end points coordinates, shape (#los, 3)
    """
    directions = generate_directions(num, fov, axis, elong)
    start = np.full_like(directions, pinhole)
    end = directions + pinhole
    return np.array((start, end))


def rot_v(points, angle):
    """
    Rotates given points in vertical direction, that is about horizontal y axis perpendicular to r/x.

    Should be done before horizontal rotation.

    Parameters
    ----------
    points : numpy.ndarray 
        3D coordinates of points to be rotated with shape (#points, 3)
    angle : float
        angle of rotation in radians

    Returns
    --------
    numpy.ndarray
        array of rotated points
    """
    s = np.sin(angle)
    c = np.cos(angle)
    mat = np.array(((c, 0, s),
                    (0, 1, 0),
                    (-s, 0, c)))
    rpoints = points.dot(mat)
    return rpoints


def rot_h(points, angle):
    """
    Rotates given points in horizontal direction, that is about vertical axis z.

    Parameters
    -----------
    points : numpy.ndarray 
        3D coordinates of points to be rotated with shape (#points, 3)
    angle : float
        angle of rotation in radians

    Returns
    --------
    numpy.ndarray
        array of rotated points
    """
    s = np.sin(angle)
    c = np.cos(angle)
    mat = np.array(((c, s, 0),
                    (-s, c, 0),
                    (0, 0, 1)))
    rpoints = points.dot(mat)
    return rpoints



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Module with tkinter apps for computing tomography reconstructions.
"""
import sys
import tkinter as tk
from pathlib import Path
from tkinter import ttk
from tkinter import filedialog

import h5py
import numpy as np

from .widgets import MasterControls
from .widgets_tomo import DiagGraphWindow, GraphPower, GraphRgb
from .roi import GraphRectangles, GraphContours
from .profiler import GraphProfiler
from .merfer import GraphMerfer
from .menus import AnalyzerMenu, ExpandMenu, SelectorMenu


class AnalyzerApp(ttk.Frame):
    def __init__(self, master=None):
        master.rowconfigure(0, weight=1)
        master.columnconfigure(0, weight=1)
        master.title('Results analyzer')
        master.bind('<Return>', self.do)
        ttk.Frame.__init__(self, master, padding=(3, 3, 3, 3))
        self.grid()
        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, weight=1)
        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)
        self.menu = AnalyzerMenu(self)
        self.menu.grid(row=0, column=0, columnspan=2, sticky='NSWE')
        self.but = ttk.Button(self, text='do', command=self.do)
        self.but.grid(row=2, column=0, sticky='W')
        self.graf_bol = tk.Frame()
        self.graf_bol.destroy()
        self.graf_ntr = tk.Frame()
        self.graf_ntr.destroy()
        self.graf_sxr = tk.Frame()
        self.graf_sxr.destroy()
        self.graf_rgb = tk.Frame()
        self.graf_rgb.destroy()
        self.graf_rec = tk.Frame()
        self.graf_rec.destroy()
        self.graf_prf = tk.Frame()
        self.graf_prf.destroy()
        self.graf_pwr = tk.Frame()
        self.graf_pwr.destroy()
        self.graf_cnt = tk.Frame()
        self.graf_cnt.destroy()
        self.graf_mrf = tk.Frame()
        self.graf_mrf.destroy()
        self.mc = tk.Frame()
        self.mc.destroy()
        self.exp = None
        self.data = {}
        self.shot = 0
        self.resolution = None
        self.form = None

    def do(self, *args):
        """
        Loads data from files and creates requested graphs and analytical widgets.
        """
        # TODO hdf loading or selection of file and suffix detection
        stime = float(self.menu.stimevar.get())
        etime = float(self.menu.etimevar.get())
        self.data['tind'] = {}
        self.data['shot'] = self.menu.shotvar.get()
        self.shot = self.data['shot']
        self.resolution = self.menu.rescbox.get()
        names = {'bol': 'bolometers',
                    'ntr': 'neutrons',
                    'sxr': 'sxr',
                    }
        switches = {
            'bol': self.menu.bolvar,
            'ntr': self.menu.ntrvar,
            'sxr': self.menu.sxrvar,
        }
        tokname = self.menu.tokvar.get().capitalize()
        path = Path('Tomodata/{}/Output'.format(tokname))
        if tokname == 'Jet':
            for d in names:
                name = names[d]
                if switches[d].get():
                    try:
                        name = '{}/tomoresults_{}_{}'.format(name.capitalize(), self.shot, name)
                        floc = str(path / name)
                        res = self.resolution.split('x')
                        self.data[d] = {}
                        with h5py.File(floc, 'r') as fset:
                            tvec = fset['time'][...]
                            self.data[d]['tvec'] = tvec
                            self.data[d]['R'] = fset['r'][...]
                            self.data[d]['Z'] = fset['z'][...]
                            resh = [len(tvec), int(res[0]), int(res[1])]
                            tmp = fset['recs'][...]
                            tmp = tmp.reshape(resh)
                            self.data[d]['Recs'] = tmp
                            tmp = fset['ptot'][...]
                            tmp = tmp.values.reshape(len(tvec))
                            self.data[d]['TOPO'] = tmp
                            tmp = fset['used_data'][...]
                            tmp = tmp.values.reshape(-1, len(tvec))
                            self.data[d]['meli'] = tmp
                            tmp = fset['retrofit'][...]
                            tmp = tmp.values.reshape(-1, len(tvec))
                            self.data[d]['bcli'] = tmp
                            self.data['tind'][d] = (stime <= self.data[d]['tvec']) * (self.data[d]['tvec'] <= etime)
                    except Exception as e:
                        print('Unable to load {} data.'.format(names[d]))
                        print(e)
        elif tokname == 'Compass':
            name = 'axuv'
            d = 'bol'
            name = '{}/tomoresults_{}_{}'.format(name.capitalize(), self.shot, name)
            try:
                floc = str(path / name)
                res = self.resolution.split('x')
                self.data[d] = dict()
                with h5py.File(floc, 'r') as fset:
                    tvec = fset['time'][...]
                    self.data[d]['tvec'] = tvec
                    self.data[d]['R'] = fset['r'][...]
                    self.data[d]['Z'] = fset['z'][...]
                    resh = [len(tvec), int(res[0]), int(res[1])]
                    tmp = fset['recs'][...]
                    tmp = tmp.reshape(resh)
                    self.data[d]['Recs'] = tmp
                    tmp = fset['ptot'][...]
                    tmp = tmp.values.reshape(len(tvec))
                    self.data[d]['TOPO'] = tmp
                    tmp = fset['used_data'][...]
                    tmp = tmp.values.reshape(-1, len(tvec))
                    self.data[d]['meli'] = tmp
                    tmp = fset['retrofit'][...]
                    tmp = tmp.values.reshape(-1, len(tvec))
                    self.data[d]['bcli'] = tmp
                    self.data['tind'][d] = (stime <= self.data[d]['tvec']) * (self.data[d]['tvec'] <= etime)
            except Exception as e:
                print('Unable to load {} data.'.format(name))
                print(e)
        else:
            print('Unrecognized tokamak name {}'.format(tokname))
        mpath = Path('Tomodata/Jet/Magnetics/')
        try:
            # TODO lims selector
            lims = '1.8-4.0--1.9-2.1'
            mname = 'Magnetic_Field_{shot}_{res}-{lims}.npz'.format(
                    shot=self.shot, res=self.resolution, lims=lims)
            self.data['mag'] = np.load(mpath / mname)
            mtvec = self.data['mag']['tflux']
            self.data['mtind'] = np.where((stime <= mtvec) & (mtvec <= etime))[0]
        except IOError as e:
            print('Could not load magnetics data: {}'.format(e))
        if self.menu.bolvar.get() and self.graf_bol.winfo_exists():
            self.graf_bol.update_data(self.data)
        else:
            self.graf_bol.destroy()
        if self.menu.ntrvar.get() and self.graf_ntr.winfo_exists():
            self.graf_ntr.update_data(self.data)
        else:
            self.graf_ntr.destroy()
        if self.menu.sxrvar.get() and self.graf_sxr.winfo_exists():
            self.graf_sxr.update_data(self.data)
        else:
            self.graf_sxr.destroy()
        if self.menu.rgbvar.get() or self.graf_rgb.winfo_exists():
            self.make_rgb()
        if self.menu.recvar.get() or self.graf_rec.winfo_exists():
            self.make_rec()
        if self.menu.prfvar.get() or self.graf_prf.winfo_exists():
            self.make_prf()
        if self.menu.pwrvar.get() or self.graf_pwr.winfo_exists():
            self.make_pwr()
        if self.menu.cntvar.get() or self.graf_cnt.winfo_exists():
            self.make_cnt()
        self.expand()

    def make_bol(self):
        """
        Makes Bolometry graph window or updates data in existing window.
        """
        if not self.graf_bol.winfo_exists():
            self.graf_bol = DiagGraphWindow(self, data=self.data, title='Bolometry')
        else:
            self.graf_bol.update_data(self.data)
        self.graf_bol.plot(0)

    def make_ntr(self):
        """
        Makes neutron graph window or updates data in existing window.
        """
        tind = self.data['tind']['ntr']
        tvec = self.data['ntr']['tvec'][tind]
        if not self.graf_ntr.winfo_exists():
            self.graf_ntr = DiagGraphWindow(self, data=self.data, vals=tvec, title='Neutrons')
        else:
            self.graf_ntr.update_data(self.data, vals=tvec)
        self.graf_ntr.plot(0)

    def make_sxr(self):
        """
        Makes Soft X-Ray graph window or updates data in existing window.
        """
        tind = self.data['tind']['sxr']
        tvec = self.data['sxr']['tvec'][tind]
        if not self.graf_sxr.winfo_exists():
            self.graf_sxr = DiagGraphWindow(self, data=self.data,
                                            vals=tvec, title='SXR')
        else:
            self.graf_sxr.update_data(self.data, vals=tvec)
        self.graf_sxr.plot(0)

    def make_rgb(self):
        """
        Makes RGB graph window or updates data in existing RGB graph window.
        """
        if not self.graf_rgb.winfo_exists():
            self.graf_rgb = GraphRgb(self, data=self.data)
        else:
            self.graf_rgb.update_data(self.data)
        self.graf_rgb.plot(0)

    def make_rec(self):
        """
        Makes rectangles graph window or updates data in existing window.
        """
        if not self.graf_rec.winfo_exists():
            self.graf_rec = GraphRectangles(self, data=self.data)
        else:
            self.graf_rec.update_data(self.data)
        self.graf_rec.plot(0)

    def make_prf(self):
        """
        Makes profiler graph window or updates data in existing window.
        """
        if not self.graf_prf.winfo_exists():
            self.graf_prf = GraphProfiler(self, data=self.data)
        else:
            self.graf_prf.update_data(self.data)
        self.graf_prf.plot(0)

    def make_pwr(self):
        """
        Makes power graph window or updates data in existing window.
        """
        if not self.graf_pwr.winfo_exists():
            self.graf_pwr = GraphPower(self, self.data)
        else:
            self.graf_pwr.update_data(self.data, )
    #        self.graf_pwr.plot(0)
    
    def make_mrf(self):
        """
        Makes Merfer graph window or updates data in existing RGB graph window.
        """
        if not self.graf_mrf.winfo_exists():
            self.graf_mrf = GraphMerfer(self, data=self.data)
        else:
            self.graf_mrf.update_data(self.data)
        self.graf_mrf.plot(0)

    def make_cnt(self):
        """
        Makes contours graph window or updates data in existing window.
        """
        if not self.graf_cnt.winfo_exists():
            self.graf_cnt = GraphContours(self, data=self.data)
        else:
            self.graf_cnt.update_data(self.data)

    def get_ctrl(self):
        """
        Checks for existing slave graphical windows and returns them in list

        Returns
        -------
        control : array
            Contains all slave graphs widgets.
        """
        control = []
        slaves = [self.graf_bol, self.graf_ntr, self.graf_sxr,
                  self.graf_rgb, self.graf_mrf]
        for i in slaves:
            if i.winfo_exists():
                control.append(i)
        return control

    def make_mc(self):
        """
        Makes Master controls widget for slave graphs.
        """
        control = self.get_ctrl()
        if len(control) > 0:
            tind = control[0].vals
            slices = len(tind)
            kw = {'vals': tind}
        else:
            kw = {}
            slices = 1
        if self.mc.winfo_exists():
            self.mc.ctrl = control
            self.mc.set_slices(slices, **kw)
        else:
            self.mc = MasterControls(self, slices, *control, **kw)

    def expand(self, **kw):
        """
        Executed after main 'Do' method.
        Makes buttons for recreating closed graph windows.
        """
        self.exp = ExpandMenu(self, padding=(3, 3, 12, 12), relief='solid')
        self.exp.grid(row=1, column=0, columnspan=3, sticky='NSWE')


class SelectorApp(AnalyzerApp):
    """
    Manually select hdf files to be used for comparative analysis.
    """
    def __init__(self, master=None):
        AnalyzerApp.__init__(self, master)
        self.menu.destroy()
        self.menu = SelectorMenu(self)
        self.menu.grid(row=0, column=0, columnspan=2, sticky='NSWE')
        # TODO
        # self.menu.var3.set('Output/mhd/tomoresults_6071_sxr')
        # self.menu.varm.set('Output/mhd/mag_mhd')
        self.fnames = list()

    def select1(self):
        nm = filedialog.askopenfilename(parent=self)
        self.menu.var1.set(nm)

    def select2(self):
        nm = filedialog.askopenfilename(parent=self)
        self.menu.var2.set(nm)

    def select3(self):
        nm = filedialog.askopenfilename(parent=self)
        self.menu.var3.set(nm)

    def selectm(self):
        nm = filedialog.askopenfilename(parent=self)
        self.menu.varm.set(nm)

    def do(self, *args, **kwargs):
        self.data = {}
        self.fnames = list()
        strvars = (self.menu.var1, self.menu.var2, self.menu.var3)
        stime = float(self.menu.stimevar.get())
        etime = float(self.menu.etimevar.get())
        self.data['tind'] = {}
        self.data['errors'] = {}
        for var in strvars:
            self.fnames.append(var.get())
        # TODO making dict from hdf as method of Tomoset class
        # for i, floc in enumerate(self.fnames):
        names = ('bol', 'ntr', 'sxr')
        for a in range(3):
            floc = self.fnames[a]
            d = names[a]
            self.data[d] = dict()
            if floc:
                with h5py.File(floc, 'r') as fset:
                    tvec = fset['time'][...]
                    self.data[d]['tvec'] = tvec
                    self.data[d]['R'] = fset['r'][...]
                    self.data[d]['Z'] = fset['z'][...]
                    # TODO change recs to values
                    self.data[d]['Recs'] = fset['recs'][...]
                    tmp = fset['ptot'][...]
                    tmp = tmp.reshape(len(tvec))
                    self.data[d]['TOPO'] = tmp
                    tmp = fset['used_data'][...]
                    tmp = tmp.reshape(-1, len(tvec))
                    self.data[d]['meli'] = tmp
                    tmp = fset['retrofit'][...]
                    tmp = tmp.reshape(-1, len(tvec))
                    self.data[d]['bcli'] = tmp
                    try:
                        self.data['errors'][d] = fset['errors'][...]
                    except KeyError:
                        pass
                    # TODO load only relevant time interval
                    self.data['tind'][d] = (stime <= self.data[d]['tvec']) * (self.data[d]['tvec'] <= etime)
                # TODO data loading and widget creation
                # TODO fnames to Path and check availability
        mloc = self.menu.varm.get()
        if mloc:
            magfield = np.load(mloc)
            self.data['mag'] = {}
            self.data['mag']['psiRz'] = magfield['values']
            self.data['mag']['tflux'] = magfield['time']
            mtvec = magfield['time']
            self.data['mtind'] = (stime <= mtvec) * (mtvec <= etime)

        self.exp = ttk.Frame(self)
        self.exp.grid(row=1, column=0, columnspan=3, sticky='NSWE')
        self.explbl = ttk.Label(self.exp, text='Create window:')
        self.explbl.grid(row=0, column=0, columnspan=2, sticky='W')
        #        for i in range(6):
        #            self.columnconfigure(i, weight = 1)
        bwid = 7
        self.bolbtn = ttk.Button(self.exp, text='bol', width=bwid, command=self.make_bol)
        self.bolbtn.grid(row=1, column=0)
        if not self.menu.var1.get():
            self.bolbtn.state(['disabled'])
        self.ntrbtn = ttk.Button(self.exp, text='ntr', width=bwid, command=self.make_ntr)
        self.ntrbtn.grid(row=1, column=1)
        if not self.menu.var2.get():
            self.ntrbtn.state(['disabled'])
        self.sxrbtn = ttk.Button(self.exp, text='sxr', width=bwid, command=self.make_sxr)
        self.sxrbtn.grid(row=1, column=2)
        if not self.menu.var3.get():
            self.sxrbtn.state(['disabled'])
        self.mrfbtn = ttk.Button(self.exp, text='mrf', width=bwid, command=self.make_mrf)
        self.mrfbtn.grid(row=1, column=4)
        self.rgbbtn = ttk.Button(self.exp, text='rgb', width=bwid, command=self.make_rgb)
        self.rgbbtn.grid(row=2, column=0)
        self.pwrbtn = ttk.Button(self.exp, text='pwr', width=bwid, command=self.make_pwr)
        self.pwrbtn.grid(row=2, column=1)
        self.recbtn = ttk.Button(self.exp, text='rec', width=bwid, command=self.make_rec)
        self.recbtn.grid(row=2, column=2)
        self.prfbtn = ttk.Button(self.exp, text='prf', width=bwid, command=self.make_prf)
        self.prfbtn.grid(row=2, column=3)
        self.cntbtn = ttk.Button(self.exp, text='cnt', width=bwid, command=self.make_cnt)
        self.cntbtn.grid(row=2, column=4)
        self.masterbtn = ttk.Button(self.exp, text='master', command=self.make_mc, width=bwid + 3)
        self.masterbtn.grid(row=3, column=0, columnspan=2, sticky='W')

    def make_mrf(self):
        if not self.graf_mrf.winfo_exists():
            self.graf_mrf = GraphMerfer(self, data=self.data, errors=self.data['errors'])
        else:
            self.graf_mrf.update_data(self.data, err=self.data['errors'])
        self.graf_mrf.plot(0)


def launcher(app='analyzer'):
    """
    Launches specified application and ensures stdout.

    Parameters
    ----------
    app : string
        Specifies which app should be launched. Legal options are
        and 'analyzer', 'a'. These are not case sensitive.
    """
    root = tk.Tk()
    upp = app.upper()
    if upp in ['A', 'ANALYZER']:
        app = AnalyzerApp(root)
    elif upp in ['S', 'SELECTOR']:
        app = SelectorApp(root)
    wdef = sys.stdout.write
    try:
        app.mainloop()
    finally:
        sys.stdout.write = wdef



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Contains inversion class for Biorthogonal Basis Decomposition Algorithm proposed by J. Cavlier. 
It is a simplified form of wavelet-vaguelette decomposition algorithm by R. Nguyen van Yen

.. [BOB1] Jordan Cavalier et al., Nucl. Fusion 59 (2019): 056025
.. [BOB2] R. Nguyen van Yen et al., Nucl. Fusion 52 (2011): 013005
"""
import warnings

import numpy as np
import scipy.sparse as sparse


class Bob(object):
    """
    BiOrthogonal Basis decomposition

    Attributes
    ----------
    basis : array_like
        :math:`\mathbf{b}_i` basis vectors of reconstruction plane
    dec_mat : scipy.sparse.csr_matrix
        :math:`\hat{\mathbf{e}}_i` decomposed matrix used to transform image into reconstruction plane
    dec_mat_normed : scipy.sparse.csr_matrix
        normalised decomposed matrix
    """

    def __init__(self, dec_mat=None, basis=None):
        """
        Parameters
        ----------
        dec_mat : scipy.sparse.csr_matrix, optional
            previously decomposed matrix, avoids recomputation of decomposition when provided
        basis : array_like, optional
            A set of basis vectors used for decomposition
        """
        # TODO one parameter holding both dec_mat and basis?
        super().__init__()
        self.basis = basis
        self.dec_mat = dec_mat
        self.dec_mat_normed = None
        return

    def decompose(self, gmat, basis, reg_factor=0, solver_kw: dict=None):
        """
        Decomposes the geometry matrix using basis vectors

        Parameters
        ----------
        gmat : scipy.sparse.csr_matrix
            geometry/contribution matrix
        basis : sparse matrix
            matrix with basis vectors
        reg_factor : float, optional
            regularisation factor passed to cholesky decomposition
            determines weight of regularisation by identity matrix relatively to arbitrary matrix maximum value
        solver_kw : dict
            keyword parameters passed to the solver function
        """
        solver_kw = solver_kw or {}
        if gmat.shape[0] < gmat.shape[1]:
            warnings.warn('Biorthogonal algorithm requires more lines of sights than nodes in reconstruction plane to run reliably')
        self.basis = basis
        image_base = gmat.dot(self.basis)  # e_i previously known as chi, gmat in basis
        a = image_base.T.dot(image_base).toarray()  # symmetrized geometry matrix in basis
        if reg_factor:
            a = a + a.max() * reg_factor * np.eye(*a.shape)
        b = np.eye(gmat.shape[1])
        res = np.linalg.lstsq(a, b, **solver_kw)
        c = sparse.csr_matrix(res[0])  # coefficient matrix
        self.dec_mat = image_base.dot(c)  # \hat{e}_i previously known as xi, decomposed matrix
        return

    def __call__(self, data, gmat=None, thresholding=None, **kw):
        """
        Decomposes geometry matrix and projects images

        Parameters
        ----------
        data : numpy.ndarray
            contains signals with shape (# channels, # time slices)
        gmat : scipy.sparse.csr_matrix
            geometry matrix
        thresholding : float, optional
            not implemented
        
        Returns
        -------
        numpy.ndarray
            inversion results with shape (# nodes, # time slices)
        """
        # TODO transpose data and res?
        if thresholding is not None:
            warnings.warn('Thresholding not implemented to call method. Ignoring.')
        if self.dec_mat is None:
            if gmat is None:
                raise ValueError('Gmat must be provided for decomposition')
            else:
                self.decompose(gmat)
        coeffs = self.dec_mat.T.dot(data)  # coordinates in reconstruction basis
        res = self.basis.dot(coeffs)  # result in node basis
        return res

    # TODO remove? remove normalised parameter?
    def save_decomposition(self, floc, normalised=False):
        """
        Saves 

        Parameters
        ----------
        floc : str or pathlib.Path
            file location with name
        normalised : bool, optional
            selects whether to save normalised matrix, by default False
        """
        floc = str(floc)
        if normalised:
            sparse.save_npz(floc, self.dec_mat_normed, compressed=True)
        else:
            sparse.save_npz(floc, self.dec_mat, compressed=True)

    def normalise(self, precision=1e-6):
        """
        Computes normalised decomposition matrix.

        Parameters
        ----------
        precision : float, optional
            neglects decomposition matrix rows with lower norm, by default 1e-6
        """
        image_base_adj = self.dec_mat
        kappa = sparse.linalg.norm(image_base_adj, axis=0)
        idx = kappa > precision
        norms = np.zeros(kappa.size)
        norms[idx] = (1 / kappa[idx])
        xi_norm = image_base_adj.multiply(sparse.csr_matrix(norms))
        self.dec_mat_normed = xi_norm

    # TODO replace image by reconstruction?
    # TODO return mask only?
    def thresholding(self, image, c: int, precision: float=1e-6, conv: float=1e-9):
        """
        Applies thresholding method to provided image.

        Parameters
        ----------
        image : numpy.ndarray
            flattened image with shape (#pixels, 1)
        c : int
            thresholding sensitivity constant
        precision : float, optional
            normalisation precision, by default 1e-6
        conv : float, optional
            thresholding convergence limit

        Returns
        -------
        numpy.ndarray

        Raises
        ------
        RuntimeError
            If thresholding is called before decomposition of geometry matrix
        """
        if self.dec_mat is None:
            raise RuntimeError('Decomposition must be computed prior to thresholding.')
        if self.dec_mat_normed is None:
            self.normalise(precision)
        temp = sparse.csr_matrix(image)
        a = np.abs(self.dec_mat_normed.T.dot(temp).toarray())

        threshold_2 = np.sqrt(c**2 / a.size * a.T.dot(a))
        threshold_1 = 0
        while np.abs(threshold_1-threshold_2) >= conv:
            threshold_1 = threshold_2
            a_temp = a[a <= threshold_1]
            threshold_2 = np.sqrt(c**2 / a_temp.size * a_temp.T.dot(a_temp))

        out = self.dec_mat.T.dot(image)
        out = self.basis.dot(out)
        out[a < threshold_2] = 0
        return out


class SimpleBob(Bob):
    """
    Automatically creates simple one node basis as an sparse identity matrix

    .. deprecated:: 1.1
    """

    def __init__(self, dec_mat=None, basis=None):
        warnings.warn('SimpleBob is deprecated since v1.1', DeprecationWarning)
        super().__init__(dec_mat, basis)
    
    def decompose(self, gmat, basis=None):
        if basis is not None:
            warnings.warn('Ignoring basis input')
        basis = sparse.eye(gmat.shape[1])
        return super().decompose(gmat, basis)


class SparseBob(Bob):
    """
    Biorthogonal Basis Decomposition optimized for sparse matrices using inverse matrix calculation.
    """

    def decompose(self, gmat, basis, reg_factor=0, solver_kw=None):
        if solver_kw is not None:
            raise TypeError('scipy.sparse.linalg.inv does not take any keywords')
        if gmat.shape[0] < gmat.shape[1]:
            warnings.warn('Biorthogonal algorithm requires more '
            'lines of sights than nodes in reconstruction plane to run reliably')
        self.basis = basis
        image_base = gmat.dot(self.basis)  # chi
        a = image_base.T.dot(image_base)
        if reg_factor:
            a = a + a.max() * reg_factor * sparse.eye(*a.shape)
        try:
            c = sparse.linalg.inv(a)
        except RuntimeError:
            raise ValueError('Singular symmetrized matrix factor. Try increasing regularisation factor.')
        self.dec_mat = image_base.dot(c)  # xi
        return


class CholmodBob(Bob):
    """
    Decomposition optimized for sparse matrices using Cholesky decomposition

    Uses sksparse.cholmod.cholesky to solve the decomposition
    Requires positive definite symmetrized geometry matrix in reconstruction plane basis.
    """

    def decompose(self, gmat, basis, reg_factor=1e-3, solver_kw=None):
        """
        Decomposes geometry matrix using Cholesky decomposition and projects images

        Parameters
        ----------
        gmat : scipy.sparse.csr_matrix
            geometry/contribution matrix
        basis : sparse matrix
            matrix with basis vectors
        reg_factor : float, optional
            regularisation factor passed to cholesky decomposition
            determines weight of regularisation by identity matrix relatively to arbitrary matrix maximum value
        solver_kw : dict
            keyword parameters passed to the solver function
        """
        solver_kw = solver_kw or {}
        from sksparse.cholmod import cholesky, CholmodNotPositiveDefiniteError
        if gmat.shape[0] < gmat.shape[1]:
            warnings.warn('Biorthogonal algorithm can be prone to failure if there are more '
            'lines of sights than nodes in reconstruction plane')
        self.basis = basis
        image_base = gmat.dot(self.basis)  # chi
        a = image_base.T.dot(image_base)
        try:
            factor = cholesky(a, a.max()*reg_factor, **solver_kw)
        except CholmodNotPositiveDefiniteError:
            raise ValueError('Symmetrized matrix was not positive definite. Try increasing regularisation factor.')
        b = sparse.csc_matrix(np.eye(gmat.shape[1]))
        c = factor(b)
        self.dec_mat = image_base.dot(c)  # xi
        return



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Minimum Fisher Regularisation implemented based on articles by M. Anton and J. Mlynar

.. [MFR1] M. Anton et al., "X-ray tomography on the TCV tokamak.", Plasma Phys. Control. Fusion  38.11 (1996): 1849.

.. [MFR2] J. Mlynar et al., "Current research into applications of tomography for fusion diagnostics." J. Fusion Energy 38.3 (2019): 458-466
"""
import time
from warnings import warn

import numpy as np
import scipy.sparse as sparse
from scipy.sparse.linalg import spsolve
from scipy.optimize import minimize_scalar


class Mfr(object):
    r"""
    Inverses provided data using Minimum Fisher Regularization scheme.

    Attributes
    ----------
    last_chi : float
        person test result of the last computed inversion
    logalpha : float
        :math:`\log(\alpha)` logarithm of regularisation parameter
    gmat : numpy.ndarray
        geometry matrix with shape (#channels, #nodes)
    signal : numpy.ndarray
    gdg : numpy.ndarray
       :math:`\mathbf{T}^T \cdot \mathbf{T}` symmetrised geometry matrix, shape (#nodes, #nodes)
    gdsig : numpy.ndarray
        right side of regularisation scheme :math:`\mathbf{T}^T \cdot \mathbf{f}`
    """
    def __init__(self):
        super().__init__()
        self._gmat = None
        self._signal = None
        self._gdg = None
        self._gdsig = None

    @staticmethod
    def solve(a, b):
        r"""
        Finds solution of :math:`\mathbf{Ax}=\mathbf{b}` using scipy.sparse.linalg.spsolve
        """
        return spsolve(a, b)

    def _test_regularization(self, logalpha, objective):
        """
        Function passed to minimisation function used for finding regularisation parameter value.

        Inverses signals using given regularisation parameter and computes chi2 test

        Stores pearson test result in attribute last_chi.

        Parameters
        ----------
        logalpha : float
            natural logarithm of regularisation parameter
        objective : numpy.ndarray
            a discretised objective functional also called regularisation or smoothing matrix, with shape (#nodes, #nodes)
            matrix created from smoothing matrix and results from previous MFR step

        Returns
        -------
        abs(chi2 - 1) : float
            1D Euclidean distance from ideal Pearson test result
        """
        alpha = 10 ** logalpha
        mod_mat = self._gdg + alpha * objective
        g = self.solve(mod_mat, self._gdsig)
        chi2 = self._pearson_test(g)
        return abs(chi2 - 1)

    def invert(self, signals, gmat, derivatives, w_factor=None, mfi_num=3, bounds=(-15, 0), iter_max=15, w_max=1,
               aniso=0, tolerance=0.05):
        """
        Inverses normalised signals using `mfi_num` Fisher Information cycles each with 
        `iter_max` steps of regularisation parameter optimisation.

        See regularisation_matrix documentation for more information about derivative matrix formats.

        Parameters
        ----------
        signals : numpy.ndarray
            error normalised signals on detector channels
        gmat : scipy.sparse.csr_matrix
            geometry matrix normalised by estimated errors
        derivatives : list
            list of tuples containing pairs of sparse derivatives matrices
        w_factor : numpy.ndarray, optional
            weight matrix multipliers with shape (#nodes, )
        mfi_num : float, optional
            number of MFR iterations
        bounds : tuple of two floats, optional
            exponent values for bounds of regularisation parameter alpha
        iter_max : float, optional
            maximum number of root finding iterations
        w_max : float, optional
            value used in weigh matrix for zero or negative nodes
        aniso : float, optional
            Determines anisotropy of derivative matrices
        tolerance : float, optional
            End MFI loop if residuum reaches interval (1-tolerance; 1+tolerance), by default 0.05

        Returns
        -------
        g : numpy.ndarray
            vector with nodes emissivities, shape (#nodes, )
        stats : dict
            inversion statistics

                chi : float
                    Pearson test value of final result
                niter : list of int
                    numbers of iterations taken in each MFI loop
                logalpha : float
                    logarithm of final regularisation parameter

        See Also
        --------
        regularisation_matrix
        """
        ela = time.time()
        self._signal = signals
        self._gmat = gmat
        self._gdg = gmat.T.dot(gmat)
        self._gdsig = gmat.T.dot(signals)
        npix = gmat.shape[1]
        g = np.ones(npix)
        mfi_count = 0
        niter = []
        while mfi_count < mfi_num:
            # MFI loop searching for ideal value of regularisation parameter
            w = 1 / g
            w[w < 0] = w_max
            w = sparse.diags(w)
            if w_factor is not None:
                w = w * sparse.diags(w_factor)
            objective = self.regularisation_matrix(derivatives, w, aniso)
            res = minimize_scalar(self._test_regularization,
                                  method='bounded',
                                  bounds=bounds,
                                  args=objective,
                                  options={'maxiter': iter_max, 'xatol': tolerance},
                                  )
            if res.status == 1:
                warn('Maximum number of iteration in MFI loop reached. Consider increasing iter_max.')
            niter.append(res.nfev)
            # TODO write custom optimisation routine to avoid recalculating optimal solution?
            m = self._gdg + 10 ** res.x * objective
            g = self.solve(m, self._gdsig)
            mfi_count += 1
        logalpha = res.x
        last_chi = self._pearson_test(g)
        ela = time.time() - ela
        print('last chi^2 = {:.4f}, time: {:.2f} s'.format(last_chi, ela))
        stats = dict(chi=last_chi, logalpha=logalpha, niter=niter)
        return g, stats

    def smoothing_mat(self, w, derivatives, danis):
        """
        .. deprecated :: 1.1
            Use :meth:`regularisation_matrix`  
        """
        warn('Smoothing_mat deprecated by regularisation_matrix method since v1.1.', DeprecationWarning)
        self.regularisation_matrix(derivatives, w, danis)

    def regularisation_matrix(self, derivatives, w, aniso=0):
        """
        Computes nonlinear regularisation matrix from provided derivative matrices and node weight factors 
        determined by emissivity from previous iteration of the inversion loop.
        
        Anisotropic coefficients are computed using sigmoid function.

        Multiple pairs of derivatives matrices computed by different numerical scheme can be used. Each pair should
        contain derivatives in each locally orthogonal direction.

        Parameters
        ----------
        w : scipy.sparse.dia.dia_matrix
            (#nodes, #nodes) diagonal matrix with pixel weight factors
        derivatives : list of scipy.sparse.csc_matrix pairs
            contains sparse derivative matrices, each pair contains derivatives in both locally orthogonal coordinates
        aniso : float
            anistropic factor, positive values make derivatives along first coordinate more significant

        Returns
        -------
        scipy.sparse.csc_matrix
        """
        # sigmoid function
        w1 = 1 / (1 + np.exp(-aniso))
        w2 = 1 / (1 + np.exp(aniso))
        h1 = 0
        h2 = 0
        for pair in derivatives:
            h1 = h1 + pair[0].T.dot(w).dot(pair[0])
            h2 = h2 + pair[1].T.dot(w).dot(pair[1])
        smooth = w1 * h1 + w2 * h2
        return smooth

    def _pearson_test(self, g):
        r"""
        Computes retrofit and residuum :math:`\chi^2` using pearson test

        .. math ::
            \chi^2 = \frac{1}{M} \sum_{i}^{M} \left(\tilde{\mathbf{f}} - \tilde{\mathbf{T}} \cdot \mathbf{g} \right)_i^2

        Parameters
        ----------
        g : numpy.ndarray
            vector of tested emissivity

        Returns
        -------
        float
        """
        retrofit = self._gmat.dot(g)
        misfit = retrofit - self._signal
        misfit_sq = np.power(misfit, 2)
        res = np.average(misfit_sq)
        return res

    def __call__(self, data, gmat, derivatives, errors, **kwargs):
        """
        Normalises signal and geometry matrix using estimated errors of measurement and
        then executes a sequence of MFR reconstructions.

        Parameters
        ----------
        data : numpy.ndarray
            input to be inverted with shape (#channels,), (#time slices, #channels)
        gmat : sparse.csr_matrix
            shape (#channels, #nodes)
        derivatives : list
            list of tuples with derivative matrix pairs
        errors : int, float or np.ndarray
            Can have shapes (#channels, ), (#time slices,) or (#time slices, #channels)
        **kwargs : dict
            keywords passed to invert method
        
        Returns
        -------
        res : numpy.ndarray
            tomographic reconstruction, with shape (#time slices, #nodes)
        stats_list : list of dicts
            contains dicts with inversion statistics returned by invert method
        
        See also
        --------
        invert
        """
        try:
            kwargs['aniso'] = kwargs.pop('danis')
            warn('Parameter `danis` renamed to `aniso` since v1.1', DeprecationWarning)
        except KeyError:
            pass

        data_ndim = np.ndim(data)
        if data_ndim == 0:
            raise ValueError('Data must be at least an 1D array of values.')
        elif data_ndim == 1:  # flat data, assume one time slice
            data = data.reshape(1, -1)
        elif data_ndim > 2:
            raise ValueError('Data array has too many dimension: {}. Max is 2.'.format(data_ndim))

        if len(gmat.shape) != 2:
            raise ValueError('Gmat must be a 2D array or matrix')

        nslices = data.shape[0]
        nchnls = data.shape[1]
        if nchnls != gmat.shape[0]:
            raise ValueError('Different number of channels in data and gmat')
        nnodes = gmat.shape[1]

        err_ndim = np.ndim(errors)
        if err_ndim == 0:  # constant errors
            errors = np.full_like(data, errors)
        elif err_ndim == 1:  # flat errors, check for matching shape
            if errors.size == nslices:
                errors = errors.reshape(-1, 1)
            elif errors.size == nchnls:
                errors = errors.reshape(1, -1)
            else:
                raise ValueError('Incompatible shape of errors {} with the data {}'.format(errors.shape, data.shape))
            errors = errors * np.ones_like(data)  # broadcast to right shape
        elif err_ndim > 2:
            raise ValueError('Errors array has too many dimensions.')

        if errors.shape != data.shape:
            raise ValueError('Data shape {} does not match errors shape {}.'.format(data.shape, errors.shape))

        signal_nrm = data / errors

        res = np.empty((nslices, nnodes))
        stats_list = []

        for i in range(nslices):
            signal_np = signal_nrm[i, :]
            error_sp = sparse.diags(1/errors[i, :])
            gmat_nrm = error_sp.dot(gmat)
            res[i], stats = self.invert(signal_np, gmat_nrm, derivatives, **kwargs)
            stats_list.append(stats)

        return res, stats_list


class CholmodMfr(Mfr):
    """
    Uses sparse cholesky decomposition for solving the parameter optimisation task in MFI loop.
    Requires scikit sparse to be installed in order to initialize properly.

    Uses sksparse.cholmod.cholesky to solve the regularised task in parameter optimisation.
    """
    def __init__(self):
        """
        Executes standard initialization and imports sksparse.cholmod
        """
        super().__init__()
        from sksparse.cholmod import cholesky
        self.cholesky = cholesky

    def solve(self, a, b):
        r"""
        Finds solution of :math:`\mathbf{Ax}=\mathbf{b}` using sksparse.cholmod.cholesky

        Parameters
        ----------
        a : scipy.sparse.csr_matrix
            square and positive definite matrix
        b : array_like
            right hand side vector
        """
        factor = self.cholesky(a)
        return factor(b)



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Region of interest analysis widgets
"""
import os
import tkinter as tk
from tkinter import ttk

import numpy as np
from matplotlib import patches

from tomotok.core.phantoms import iso_psi
from .widgets_tomo import GraphRgb


class GraphRectangles(GraphRgb):
    """
    Based on GraphRgb class. Usefull for observing power in specified areas.
    
    Parameters
    ----------
    parent : TKInter class parent
        Default is None.
    data : dict
        Contains data, default is empty dict.
    
    Keywords
    --------
    title : str
        Name of window.
    """

    # TODO axis limits in cm
    def __init__(self, parent=None, **kw):
        #        tk.Toplevel.__init__(self, parent)
        GraphRgb.__init__(self, parent, **kw)
        if not 'title' in kw: self.title('Rectangles')
        #        xmax = self.shape[1]
        #        ymax = self.shape[0]
        #        wmax = xmax
        #        hmax = ymax
        self.ax_orig = self.ax.get_position()
        self.ax2 = self.fig.add_subplot(122)
        self.ax2.set_axis_off()

        self.pprop = ttk.Frame(self, padding=(3, 3, 3, 3))  # , relief = 'solid')
        xrow = 1
        self.pprop.grid(row=0, column=2, rowspan=2, sticky='NSWE')
        self.xvar = tk.StringVar(value=self.xmin)
        self.xsbox = tk.Spinbox(self.pprop,
                                values=tuple(self.xgrid - self.dx / 2),
                                #                                from_=0, to=xmax - 1, increment=1.,
                                textvariable=self.xvar,
                                command=self.movex, width=5)
        self.xsbox.grid(row=xrow, column=1)
        self.xlbl = ttk.Label(self.pprop, text='left')
        self.xlbl.grid(row=xrow, column=0)

        self.yvar = tk.StringVar(value=self.ymin)
        self.ysbox = tk.Spinbox(self.pprop,
                                values=tuple(self.ygrid - self.dy / 2),
                                #                                from_ = 0, to=ymax - 1, increment = 1.,
                                textvariable=self.yvar,
                                command=self.movey, width=5)
        self.ysbox.grid(row=xrow + 1, column=1)
        self.ylbl = ttk.Label(self.pprop, text='bottom')
        self.ylbl.grid(row=xrow + 1, column=0)

        self.wvar = tk.StringVar(value=self.dx)
        wvals = tuple(self.xgrid - self.xgrid[0])[:-1]
        self.wsbox = tk.Spinbox(self.pprop,
                                values=wvals,
                                #                                from_ = 1, to = wmax, increment = 1.,
                                textvariable=self.wvar,
                                command=self.movew, width=5)
        self.wsbox.grid(row=xrow + 2, column=1)
        self.wlbl = ttk.Label(self.pprop, text='width')
        self.wlbl.grid(row=xrow + 2, column=0)

        self.hvar = tk.StringVar(value=self.dy)
        hvals = tuple(self.ygrid - self.ygrid[0])[:-1]
        self.hsbox = tk.Spinbox(self.pprop,
                                values=hvals,
                                #                                from_ = 1, to = hmax, increment = 1.,
                                textvariable=self.hvar,
                                command=self.moveh, width=5)
        self.hsbox.grid(row=xrow + 3, column=1)
        self.hlbl = ttk.Label(self.pprop, text='height')
        self.hlbl.grid(row=xrow + 3, column=0)

        self.snvar = tk.StringVar(value=1)
        self.snsbox = tk.Spinbox(self.pprop, from_=1,
                                 to=min(100, len(self.vals)),
                                 increment=1,
                                 textvariable=self.snvar,
                                 width=5)
        self.snsbox.grid(row=xrow + 6, column=1)
        self.snlbl = ttk.Label(self.pprop, text='n')
        self.snlbl.grid(row=xrow + 6, column=0)
        self.snbtn = ttk.Button(self.pprop, text='Save n-th',
                                command=self.save_nth)
        self.snbtn.grid(row=xrow + 7, column=0, columnspan=2)

        self.pcount = 0
        self.pnum = tk.StringVar(value=-1)
        self.patches = []
        self.pnumlist = tk.StringVar()
        self.pnumlist.set([i + 1 for i in range(self.pcount)])
        self.pnumlbl = ttk.Label(self.pprop, text='Patch:')
        self.pnumlbl.grid(row=0, column=0)
        self.pnumcbox = ttk.Combobox(self.pprop, values=self.pnumlist,
                                     textvariable=self.pnum,
                                     state='readonly', width=5)
        self.pnumcbox.grid(row=0, column=1)

        self.addbut = ttk.Button(self.pprop, text='+',
                                 command=self.make_patch, width=4)
        self.addbut.grid(row=5, column=0)
        self.rembut = ttk.Button(self.pprop, text='-',
                                 command=self.remove_patch, width=4)
        self.rembut.grid(row=5, column=1)

        self.dobut = ttk.Button(self.pprop, text='Compute',
                                command=self.compute, width=12)
        self.dobut.grid(row=6, columnspan=2, sticky='WE')

        predef = self.get_predef()
        for i in predef:
            self.make_patch(**i)
        self.pnum.trace('w', self.pselect)
        self.pnum.set(1)

        self.style = {'bol': 'solid', 'ntr': 'dashdot', 'sxr': 'dashed'}

    def pselect(self, *args):
        """
        Changes spinbox values to selected patch when selection changes.
        """
        val = self.get_pnum()
        if val >= 0:
            x = self.ax.patches[val].get_x()
            self.xvar.set(x)
            y = self.ax.patches[val].get_y()
            self.yvar.set(y)
            w = self.ax.patches[val].get_width()
            self.wvar.set(w)
            h = self.ax.patches[val].get_height()
            self.hvar.set(h)

    def get_predef(self):
        """
        Predefined rectangles in interesting areas are defined here.
        
        Returns
        -------
        dict
            Contains coordinates of predefined rectangles
        """
        out = [
            {'x': -1, 'y': -1, 'w': -1, 'h': -1},
            #                {'x' : 25, 'y' : 37, 'w' : 15, 'h' : 30},
            #                {'x' : 40, 'y' : 37, 'w' : 10, 'h' : 30},
            # {'x' : 10, 'y' : 70, 'w' : 20, 'h' : 10},
        ]
        return out

    def get_pnum(self):
        """
        Variable pnum represents selected patch, starts from 1.
        
        Returns
        -------
        out : int
            Number of selected patch (rectangle) starting from 0. Can be used
            as idices for Axes.patches array.
        """
        out = int(self.pnum.get()) - 1
        return out

    def raise_pcount(self):
        """
        Variable pcount represents total number of patches.
        Should be called after adding a new patch.
        """
        self.pcount += 1
        val = self.pcount
        #        self.patches += [val]
        self.pnumcbox['values'] = [i + 1 for i in range(val)]

    def lower_pcount(self):
        """
        Variable pcount represents total number of patches.
        Should be called after removing a new patch.
        """
        self.pcount -= 1
        val = self.pcount
        #        self.patches += [val]
        self.pnumcbox['values'] = [i + 1 for i in range(val)]

    def make_patch(self, x=-1, y=-1, w=-1, h=-1):
        """
        Creates new patch and adds it to axes.
        """
        if x == -1:
            x = self.xmin
        if y == -1:
            y = self.ymin
        if w == -1:
            w = self.dx
        if h == -1:
            h = self.dy
        val = self.pcount
        self.ax.add_patch(patches.Rectangle((x, y), w, h, fill=0, lw=2,
                                            ec='C{}'.format(val % 10)))
        self.raise_pcount()
        self.pnum.set(val + 1)
        self.gcanvas.draw()

    def remove_patch(self):
        """
        Removes last patch from computing and from axes.
        """
        val = self.pcount - 1
        if val > 0:
            self.ax.patches[val].remove()
            self.lower_pcount()
            self.plot(self.scale.get())
            pnum = int(self.pnum.get())
            if pnum > val:
                self.pnum.set(val)

    def movex(self, *args):
        pnum = self.get_pnum()
        p = self.ax.patches[pnum]
        val = float(self.xsbox.get())
        p.set_x(val)  # +self.xmin)
        self.gcanvas.draw()

    def movey(self, *args):
        pnum = self.get_pnum()
        p = self.ax.patches[pnum]
        val = float(self.ysbox.get())
        p.set_y(val)  # +self.ymin)
        self.gcanvas.draw()

    def movew(self, *args):
        pnum = self.get_pnum()
        p = self.ax.patches[pnum]
        val = float(self.wsbox.get())
        p.set_width(val)
        self.gcanvas.draw()

    def moveh(self, *args):
        pnum = self.get_pnum()
        p = self.ax.patches[pnum]
        val = float(self.hsbox.get())
        p.set_height(val)
        self.gcanvas.draw()

    def compute(self, *args):
        style = self.style
        #        col = self.col
        self.ax.set_position([0.125, 0.125, 0.35227272727272724, 0.755])
        self.ax2.cla()
        self.ax2.set_axis_on()
        self.ax2.set_xlabel('Time [s]')
        self.ax2.set_ylabel('Pixel normalized power [-]')
        self.ax2.yaxis.tick_right()
        self.ax2.yaxis.set_label_position('right')
        count = self.pcount
        lines = {}
        for i in self.diags:
            lines[i] = []
            self.ax2.plot([], c='k', ls=style[i], label=i)
        for i in range(count):
            p = self.ax.patches[i]
            l = np.argmin(np.abs(self.xgrid - float(p.get_x())))
            b = np.argmin(np.abs(self.ygrid - float(p.get_y())))
            r = l + int(np.round(float(p.get_width()) / self.dx))
            t = b + int(np.round(float(p.get_height()) / self.dy))
            for d in self.diags:
                lines[d].append(np.sum(self.data[d][..., b:t, l:r], axis=(1, 2)))
                #                lines[d].append(np.sum(self.odata[d][b:t,l:r,...],axis = (0,1)))
                self.ax2.plot(self.tvecs[d], lines[d][i], c='C{}'.format(i), ls=style[d])
        self.ax.legend()
        self.ax2.legend()
        ind = self.scale.get()
        self.vline = self.ax2.axvline(self.vals[ind], c='k', ls='dashed', lw=1)
        y = self.ax2.get_ylim()[1] * 1.01
        x = self.vals[ind]
        self.text = self.ax2.text(x, y, str(x))
        self.gcanvas.draw()

    def save(self, fname='Graphs/'):
        if not os.path.exists(fname):
            os.makedirs(fname)
        v = self.scale.get()
        self.fig.savefig(fname +
                         '{} at {:.0f} {:.0f}-{:.0f}ms rectangles.png'.format(
                         # '{} at {:.0f} {:.0f}-{:.0f}ms rectangles.eps'.format(
                            self.shot, 1000 * self.vals[v], 1000 * self.vals[0], 1000 * self.vals[-1]))

    def update_data(self, ndata, **kw):
        GraphRgb.update_data(self, ndata, **kw)
        if self.grids[0]:
            self.xgrid = ndata[self.key]['R']
        else:
            self.xgrid = np.arange(self.shape[1])
        if self.grids[1]:
            self.ygrid = ndata[self.key]['Z']
        else:
            self.ygrid = np.arange(self.shape[0])
        if len(self.fig.axes) > 1:
            self.ax.set_position(self.ax_orig)
            self.ax2.cla()
            self.ax2.set_axis_off()
        self.ax.lines.clear()
        for i in self.diags:
            self.ax.plot([], c=self.col[i], ls='', marker='s', label=i)
        self.ax.legend()
        self.gcanvas.draw()

    def plot(self, val, *args):
        inds = self.get_rgbval(val)
        b = self.data['bol'][inds[0]]
        n = self.data['ntr'][inds[1]]
        s = self.data['sxr'][inds[2]]
        rgb = np.stack((b, n, s), axis=2)
        rgb = rgb / np.max(rgb)
        cmy = np.ones_like(rgb) - rgb
        rgb = cmy
        if self.grids[0]:
            self.ax.set_xlabel('R [m]')
        else:
            self.ax.set_xlabel('Pixel')
        if self.grids[0]:
            self.ax.set_ylabel('z [m]')
        else:
            self.ax.set_ylabel('Pixel')

        if len(self.ax.images) == 0:
            self.ax.imshow(rgb, origin='lower', extent=self.extent)
        else:
            self.ax.images[0].set_data(rgb)
        #            self.ax.images[0].set_extent(self.extent)
        for i in self.ax.collections:
            i.remove()
        if self.magvar.get():
            mval = self.get_mval(val)
            self.cont = self.ax.contour(self.mdata[mval], [1.0],
                                        extent=self.extent,
                                        # colors='w',
                                        colors='k',
                                        linewidths=1)
        if len(self.ax2.texts) > 0:
            x = self.vals[val]
            self.vline.set_xdata(x)
            self.text.set_x(x)
            self.text.set_text(str(x))
        self.gcanvas.draw()

    def save_nth(self):
        """
        Saves every nth slice, where n is defined by snvar Stringvar asociated
        with snsbox Spinbox widget.
        """
        i = 0
        n = int(self.snvar.get())
        while i < len(self.vals):
            self.go_to(i)
            self.plot(i)
            self.save(fname='Graphs/recseq/{}/'.format(self.shot))
            i += n

class GraphContours(GraphRgb):
    """
    Based on GraphRgb class. Usefull for observing power in specified areas.
    
    Parameters
    ----------
    parent : TKInter class parent
        Default is None.
    data : dict
        Contains data, default is empty dict.
    
    Keywords
    --------
    title : str
        Name of window.
    """

    def __init__(self, parent=None, **kw):
        #        tk.Toplevel.__init__(self, parent)
        GraphRgb.__init__(self, parent, **kw)
        if 'title' not in kw:
            self.title('Contours')
        self.ax_orig = self.ax.get_position()
        self.ax2 = self.fig.add_subplot(122)
        self.ax2.set_axis_off()

        self.pprop = ttk.Frame(self, padding=(3, 3, 3, 3))  # , relief = 'solid')
        xrow = 1
        self.pprop.grid(row=0, column=2, rowspan=2, sticky='NSWE')
        self.minvar = tk.StringVar(value=10)
        self.minsbox = tk.Spinbox(self.pprop, from_=0., to=2,
                                  increment=0.01, textvariable=self.minvar,
                                  command=self.move_limits, width=5)
        self.minsbox.grid(row=xrow, column=1)
        self.minlbl = ttk.Label(self.pprop, text='min')
        self.minlbl.grid(row=xrow, column=0)

        self.maxvar = tk.StringVar(value=10)
        self.maxsbox = tk.Spinbox(self.pprop, from_=0., to=2,
                                  increment=0.01, textvariable=self.maxvar,
                                  command=self.move_limits, width=5)
        self.maxsbox.grid(row=xrow + 1, column=1)
        self.maxlbl = ttk.Label(self.pprop, text='max')
        self.maxlbl.grid(row=xrow + 1, column=0)

        self.snvar = tk.StringVar(value=1)
        self.snsbox = tk.Spinbox(self.pprop, from_=1,
                                 to=min(100, len(self.vals)),
                                 increment=1,
                                 textvariable=self.snvar,
                                 width=5)
        self.snsbox.grid(row=xrow + 6, column=1)
        self.snlbl = ttk.Label(self.pprop, text='n')
        self.snlbl.grid(row=xrow + 6, column=0)
        self.snbtn = ttk.Button(self.pprop, text='Save n-th',
                                command=self.save_nth)
        self.snbtn.grid(row=xrow + 7, column=0, columnspan=2)

        self.pcount = 0
        self.pnum = tk.StringVar(value=-1)
        self.patches = []
        self.pnumlist = tk.StringVar()
        self.pnumlist.set([i + 1 for i in range(self.pcount)])
        self.pnumlbl = ttk.Label(self.pprop, text='Contours:')
        self.pnumlbl.grid(row=0, column=0)
        self.pnumcbox = ttk.Combobox(self.pprop, values=self.pnumlist,
                                     textvariable=self.pnum,
                                     state='readonly', width=5)
        self.pnumcbox.grid(row=0, column=1)

        self.addbut = ttk.Button(self.pprop, text='+',
                                 command=self.make_patch, width=4)
        self.addbut.grid(row=5, column=0)
        self.rembut = ttk.Button(self.pprop, text='-',
                                 command=self.remove_patch, width=4)
        self.rembut.grid(row=5, column=1)

        self.dobut = ttk.Button(self.pprop, text='Compute',
                                command=self.compute, width=12)
        self.dobut.grid(row=6, columnspan=2, sticky='WE')

        predef = self.get_predef()
        for i in predef:
            self.make_patch(**i)
        self.pnum.trace('w', self.pselect)
        self.pnum.set(1)
        self.style = {'bol': 'solid', 'ntr': 'dashdot', 'sxr': 'dashed'}
        self.plot(0)

    def pselect(self, *args):
        """
        Changes spinbox values to selected contours when selection changes.
        """
        val = self.get_pnum()
        if val >= 0:
            min_, max_ = self.get_levels(val)
            self.minvar.set(min_)
            self.maxvar.set(max_)

    def get_levels(self, pnum):
        """
        Gets levels from contour object. If length of levels is one,
        returns array with zero as minimum.
        
        Parameters
        ---------
        pnum : int
            Number of patch
        
        Returns
        -------
        limits : np.ndarray
            Array with minimum and maximum limited by given contours.
        """
        p = self.patches[pnum]
        limits = p.levels
        if len(limits) == 1:
            limits = np.append(0, limits)
        return limits

    def get_predef(self):
        """
        Returns predefined contours. Used for initalization of wondow.
        
        Returns
        -------
        out : dict
            Contains coordinates of predefined rectangles
        """
        out = [
            {'min_': 0.5, 'max_': 0.7},
            {'min_': 0.7, 'max_': 1.0},
            # {'x' : 10, 'y' : 70, 'w' : 20, 'h' : 10},
        ]
        return out

    def get_pnum(self):
        """
        Variable pnum represents selected contours, starts from 1.
        
        Returns
        -------
        out : int
            Number of selected contours pair starting from 0.
        """
        out = int(self.pnum.get()) - 1
        return out

    def raise_pcount(self):
        """
        Variable pcount represents total number of patches.
        Should be called after adding a new patch.
        """
        self.pcount += 1
        val = self.pcount
        #        self.patches += [val]
        self.pnumcbox['values'] = [i + 1 for i in range(val)]

    def decrease_pcount(self):
        """
        Variable pcount represents total number of patches.
        Should be called after removing a new patch.
        """
        self.pcount -= 1
        val = self.pcount
        #        self.patches += [val]
        self.pnumcbox['values'] = [i + 1 for i in range(val)]

    def add_patch(self, contours):
        """
        Adds collections created by contours to axes
        """
        for i in contours.collections.copy():
            self.ax.add_collection(i)

    def clear_patch(self, contours):
        """
        Removes collections representing given cotnours from axes.
        """
        colls = contours.collections.copy()
        for i in colls:
            self.ax.collections.remove(i)

    def make_patch(self, min_=0.5, max_=1, pnum=None):
        """
        Creates new contours and adds collections to axes.
        """
        if pnum == None:
            val = self.pcount
        else:
            val = pnum
        if float(min_) == 0.:
            ls = 'solid'
        #            lw = 3
        else:
            #            lw = [1, 3]
            #            ls = ['dotted','solid']
            ls = ['dashdot', 'solid']
        mval = self.get_mval(self.scale.get())
        cnt = self.ax.contour(self.mdata[mval], [min_, max_],
                              extent=self.extent,
                              colors=['C{}'.format(val % 10)],
                              #                              linewidths=lw,
                              linestyles=ls,
                              )
        if pnum == None:
            self.patches.append(cnt)
            self.raise_pcount()
            self.pnum.set(val + 1)
        else:
            self.patches[pnum] = cnt
        self.gcanvas.draw()

    def remove_patch(self):
        """
        Removes last patch from computing and from axes.
        """
        val = self.pcount - 1
        if val > 0:
            patch = self.patches.pop()
            self.clear_patch(patch)
            del (patch)
            self.decrease_pcount()
            self.plot(self.scale.get())
            pnum = int(self.pnum.get())
            if pnum > val:
                self.pnum.set(val)

    def move_limits(self, *args):
        pnum = self.get_pnum()
        p = self.patches[pnum]
        min_ = self.minsbox.get()
        max_ = self.maxsbox.get()
        self.clear_patch(p)
        self.make_patch(min_, max_, pnum)

    def compute(self, *args):
        style = self.style
        #        col = self.col
        self.ax.set_position([0.125, 0.125, 0.35227272727272724, 0.755])
        self.ax2.cla()
        self.ax2.set_axis_on()
        self.ax2.set_xlabel('Time [s]')
        self.ax2.set_ylabel('Pixel normalized power [-]')
        self.ax2.yaxis.tick_right()
        self.ax2.yaxis.set_label_position('right')
        count = self.pcount
        lines = {}
        for i in self.diags:
            lines[i] = []
            self.ax2.plot([], c='k', ls=style[i], label=i)

        for i in range(count):
            min_, max_ = self.get_levels(i)
            for d in self.diags:
                line = []
                for val in range(len(self.tvecs[d])):
                    mval = self.get_mval(val)
                    ge = self.mdata[mval] > min_
                    le = self.mdata[mval] <= max_
                    inds = le * ge
                    try:
                        line.append(np.sum(self.data[d][val, inds]))
                    except IndexError as e:
                        print(e)
                        self.inds = inds
                        print(inds)
                lines[d].append(line)
                self.ax2.plot(self.tvecs[d], lines[d][i], c='C{}'.format(i), ls=style[d])
        self.ax.legend()
        self.ax2.legend()
        ind = self.scale.get()
        self.vline = self.ax2.axvline(self.vals[ind], c='k',
                                      ls='dashed', lw=1)
        y = self.ax2.get_ylim()[1] * 1.01
        x = self.vals[ind]
        self.text = self.ax2.text(x, y, str(x))
        self.gcanvas.draw()

    def save(self, fname=None):
        if fname == None:
            fname = 'Graphs/'
        if not os.path.exists(fname):
            os.makedirs(fname)
        v = self.scale.get()
        self.fig.savefig(fname +
                         '{} at {:.0f} {:.0f}-{:.0f}ms contours.png'.format(
                         # '{} at {:.0f} {:.0f}-{:.0f}ms contours.eps'.format(
                             self.shot, 1000 * self.vals[v], 1000 * self.vals[0], 1000 * self.vals[-1]))

    def update_data(self, ndata, **kw):
        GraphRgb.update_data(self, ndata, **kw)
        if len(self.fig.axes) > 1:
            self.ax.set_position(self.ax_orig)
            self.ax2.cla()
            self.ax2.set_axis_off()
        self.ax.lines.clear()
        for i in self.diags:
            self.ax.plot([], c=self.col[i], ls='', marker='s', label=i)
        self.ax.legend()
        self.gcanvas.draw()

    def plot(self, val, *args):
        inds = self.get_rgbval(val)
        b = self.data['bol'][inds[0]]
        n = self.data['ntr'][inds[1]]
        s = self.data['sxr'][inds[2]]
        rgb = np.stack((b, n, s), axis=2)
        rgb = rgb / np.max(rgb)
        cmy = np.ones_like(rgb) - rgb
        rgb = cmy
        if self.grids[0]:
            self.ax.set_xlabel('R [m]')
        else:
            self.ax.set_xlabel('Pixel')
        if self.grids[0]:
            self.ax.set_ylabel('z [m]')
        else:
            self.ax.set_ylabel('Pixel')

        if len(self.ax.images) == 0:
            self.ax.imshow(rgb, origin='lower',
                           #            self.ax.imshow(cmy, origin='lower',
                           extent=self.extent,
                           )
        else:
            self.ax.images[0].set_data(rgb)
        #            self.ax.images[0].set_data(cmy)
        #            self.ax.images[0].set_extent([self.xmin, self.xmax, self.ymin, self.ymax])
        if hasattr(self, 'magcont'):
            for i in self.magcont.collections.copy():
                self.ax.collections.remove(i)
        if self.magvar.get():
            mval = self.get_mval(val)
            self.magcont = self.ax.contour(self.mdata[mval], [1.0],
                                           extent=self.extent,
                                           # colors='w',
                                           colors='k',
                                           linewidths=1)
        if len(self.ax2.texts) > 0:
            x = self.vals[val]
            self.vline.set_xdata(x)
            self.text.set_x(x)
            self.text.set_text(str(x))
        #        print(self.ax.collections)
        for i, p in enumerate(self.patches):
            min_, max_ = self.get_levels(i)
            #            print(p)
            self.clear_patch(p)
            self.make_patch(min_, max_, i)
        self.gcanvas.draw()

    def save_nth(self):
        """
        Saves every nth slice, where n is defined by snvar tk.Stringvar
        asociated with snsbox tk.Spinbox widget.
        """
        i = 0
        n = int(self.snvar.get())
        while i < len(self.vals):
            self.go_to(i)
            self.plot(i)
            self.save(fname='Graphs/cntseq/{}/'.format(self.shot))
            i += n



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
MEasured vs RetroFit comparator graphical window widget
"""
from pathlib import Path

import numpy as np

from .widgets import GraphWindow


class GraphMerfer(GraphWindow):
    """
    Widgets for comparing MEasured data with RetroFit
    
    Parameters
    ----------
    parent : TKInter widget
        Default is None.
    data : dict
        Contains data, default is empty dict.
    """
    def __init__(self, parent=None, data=None, errors=None, **kw):
        GraphWindow.__init__(self, parent, **kw)
        if data is None:
            data = {}
        if 'title' in kw:
            self.title(kw['title'])
        else:
            self.title('Merfer')
        self.colors = {'bol': 'r', 'ntr': 'g', 'sxr': 'b'}
        # self.colors = ('c', 'm', 'y')
        self.update_data(data, err=errors)

    def update_data(self, ndata, err=None, **kw):
        """
        Loads new data.
        
        Parameters
        ----------
        ndata : dict
            New data to be loaded to Graph window
        err : np.ndarray
            contains errors of measured data with shape (#chnl, #tslices)
        """
#        self.data = {}
        if err is None:
            err = {'bol': None, 'ntr': None, 'sxr': None}
        self.errors = err
        self.tvecs = {}
        self.diags = []
        try:
            self.shot = ndata['shot']
        except KeyError:
            self.shot = 0
        keys = ndata.keys()
        self.me_data = {}
        self.rf_data = {}
        if 'tind' not in keys:
            ndata['tind'] = {}
        if 'bol' in keys:
            if 'bol' in ndata['tind']:
                tind = ndata['tind']['bol']
            else:
                tind = range(len(ndata['bol']['tvec']))
            me_bol = ndata['bol']['meli'][..., tind]
            max_ = np.max(me_bol, axis=0)
            me_bol = me_bol / max_
            try:
                self.errors['bol'] = ndata['errors']['bol'] / max_
            except KeyError:
                self.errors['bol'] = None
            self.me_data['bol'] = me_bol
            rf_bol = ndata['bol']['bcli'][..., tind]
            rf_bol = rf_bol / np.max(rf_bol, axis=0)
            self.rf_data['bol'] = rf_bol
            self.tvecs['bol'] = ndata['bol']['tvec'][tind]
            self.diags.append('bol')
        if 'ntr' in keys:
            if 'ntr' in ndata['tind']:
                tind = ndata['tind']['ntr']
            else:
                tind = range(len(ndata['ntr']['tvec']))
            me_ntr = ndata['ntr']['meli'][..., tind]
            max_ = np.max(me_ntr, axis=0)
            me_ntr = me_ntr / max_
            try:
                self.errors['ntr'] = ndata['errors']['ntr'] / max_
            except KeyError:
                self.errors['ntr'] = None
            self.me_data['ntr'] = me_ntr
            rf_ntr = ndata['ntr']['bcli'][..., tind]
            rf_ntr = rf_ntr / np.max(rf_ntr, axis=0)
            self.rf_data['ntr'] = rf_ntr
            self.tvecs['ntr'] = ndata['ntr']['tvec'][tind]
            self.diags.append('ntr')
        if 'sxr' in keys:
            if 'sxr' in ndata['tind']:
                tind = ndata['tind']['sxr']
            else:
                tind = range(len(ndata['sxr']['tvec']))
            me_sxr = ndata['sxr']['meli'][..., tind]
            max_ = np.max(me_sxr, axis=0)
            me_sxr = me_sxr / max_
            try:
                self.errors['sxr'] = ndata['errors']['sxr'] / max_
            except KeyError:
                self.errors['sxr'] = None
            self.me_data['sxr'] = me_sxr
            rf_sxr = ndata['sxr']['bcli'][..., tind]
            rf_sxr = rf_sxr / np.max(rf_sxr, axis=0)
            self.rf_data['sxr'] = rf_sxr
            self.tvecs['sxr'] = ndata['sxr']['tvec'][tind]
            self.diags.append('sxr')
        self.key = self.diags[0]
        tvec = np.array(self.tvecs[self.key])
        self.set_vals(vals=tvec)
        self.plot(0)
    
    def plot(self, val, *args):
        self.ax.cla()
        for d in self.diags:
            tind = np.argmin(np.abs(self.tvecs[d] - self.vals[val]))
            chnls = np.arange(self.me_data[d].shape[0], dtype=np.float64)
            # self.ax.plot(self.me_data[d][:, tind], c='C{}'.format(i), label='Measured {}'.format(d))
            try:
                err = self.errors[d][:, tind]
            except TypeError:
                err = None
            self.ax.errorbar(chnls, self.me_data[d][:, tind], err, c=self.colors[d],
                             ls='', marker='+', capsize=3, label='Measured {}'.format(d))
            self.ax.plot(self.rf_data[d][:, tind], c=self.colors[d], alpha=0.8, label=r'Retrofit {}'.format(d))
            # sc = np.zeros(chnls.size + 2)
            # sc[0] = chnls[0] - 0.5
            # sc[-1] = chnls[-1] + 0.5
            # sc = chnls
            # self.ax.step(sc, self.rf_data[d][:, tind], c=self.colors[d],
            #              alpha=0.6, where='mid', label=r'Retrofit {}'.format(d))
            # self.ax.bar(chnls, self.rf_data[d][:, tind], edgecolor=self.colors[d], color='',
            #             label=r'Retrofit {}'.format(d))
        self.ax.legend()
        self.ax.set_xlabel('Channel')
        self.ax.set_ylabel('Normalized Power [-]')
        self.gcanvas.draw()
    
    def save(self, fname='Graphs'):
        val = self.scale.get()
        tslice = self.vals[val] * 1000
        # form = 'eps'
        form = 'png'
        name = '{} MERFer at {:.0f}ms.{}'.format(self.shot, tslice, form)
        floc = Path() / fname / name
        if not floc.parent.exists():
            floc.parent.mkdir(parents=True)
        self.fig.savefig(floc)



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Module with tkinter widgets for basic tomography analysis.
"""
import os
import tkinter as tk
from tkinter import ttk

import matplotlib.colorbar as cbar
import numpy as np

from .widgets import GraphWindow


# TODO REC normalize data?


class DiagGraphWindow(GraphWindow):
    """
    Based on GraphWindow class from widgets.py
    Used for displaying results of single diagnostic tomography

    See documentation of GraphWindow for information about other keyword arguments.
    
    Parameters
    ----------
    parent : TKInter widget
        Default is None. Used in GUI interface to specify hierarchy.
    data : dict
        Contains data, default is empty dict.
            
    **kw
        title : str {'Bolometry', 'Neutrons', 'SXR'}
            Name of graph window. If dname is not specified, title is used to 
            determin what key from provided data to load.
        dname : str {'bol', 'ntr', 'sxr'}
            Shorter string to specify diagnostic for graph window. Though not recommended, both title and dname can be used.
            If so, dname and title should always match to avoid inconsistency of title and data.
    
    See Also
    --------
    GraphWindow
    """

    def __init__(self, parent=None, data=None, **kw):
        GraphWindow.__init__(self, parent, **kw)
        if data is None:
            data = {}
        self.keys = kw.keys()
        if 'title' in self.keys:
            title = kw['title']
        else:
            title = 'Unknown diagnostic'
        if 'dname' in self.keys:
            dname = kw['dname']
        else:
            if title == 'Bolometry':
                dname = 'bol'
            elif title == 'Neutrons':
                dname = 'ntr'
            elif title == 'SXR':
                dname = 'sxr'
            else:
                dname = ''
                print('Unknown diagnostic name. Can not read data dict.',
                      'Check title or dname keyword arguments.')
        if 'shot' in self.keys:
            self.shot = kw['shot']
        elif 'shot' in data:
            self.shot = data['shot']
        else:
            self.shot = 'noshot'
        self.magvar = tk.BooleanVar(value=0)
        self.magchb = ttk.Checkbutton(self.bbox, text='Magnetics',
                                      variable=self.magvar, onvalue=True,
                                      command=self.update_graph)
        self.magchb.grid(row=self.bb_row, sticky='W')
        self.bb_row_raise()
        self.bndvar = tk.BooleanVar(value=1)
        self.bndchb = ttk.Checkbutton(self.bbox, text='Borders',
                                      variable=self.bndvar, onvalue=True,
                                      command=self.update_graph)
        self.bndchb.grid(row=self.bb_row, sticky='W')
        self.bb_row_raise()
        self.nrmvar = tk.BooleanVar(value=1)
        self.nrmchb = ttk.Checkbutton(self.bbox, text='Normalize',
                                      variable=self.nrmvar, onvalue=True,
                                      command=self.update_graph)
        self.nrmchb.grid(row=self.bb_row, sticky='W')
        self.bb_row_raise()
        self.title(title)
        self.ptitle = title
        self.dname = dname
        self.update_data(data, **kw)
        self.cax = cbar.make_axes(self.ax)[0]

    def plot(self, val, **kw):
        self.ax.cla()
        self.cax.cla()
        self.ax.set_xlabel('R [m]')
        self.ax.set_ylabel('z [m]')
        self.ax.set_title('{} at {:.4f} s'.format(self.ptitle, self.vals[val]))
        if self.nrmvar.get():
            im = self.ax.imshow(self.data[val], origin='lower', extent=[self.xmin, self.xmax, self.ymin, self.ymax],
                                vmin=self.vmin, vmax=self.vmax)
        else:
            im = self.ax.imshow(self.data[val], origin='lower', extent=[self.xmin, self.xmax, self.ymin, self.ymax])
        if self.magvar.get():
            mval = self.get_mval(val)
            self.ax.contour(self.mdata[mval], [1.0],
                            colors='w',
                            extent=[self.xmin, self.xmax, self.ymin, self.ymax],
                            linewidths=1,
                            )
        if self.bndvar.get():
            self.ax.plot(self.bnd[:, 0], self.bnd[:, 1], linewidth=2.0, color='k')
        self.fig.colorbar(im, cax=self.cax, ax=self.ax)
        self.gcanvas.draw()

    def update_data(self, ndata, **kw):
        """
        Loads new data.
        
        Parameters
        ----------
        ndata : dict
            New data to be loaded to Graph window.
        """
        if 'tind' not in ndata:
            ndata['tind'] = {}
        if self.dname in ndata['tind']:
            tind = ndata['tind'][self.dname]
        else:
            tind = range(len(ndata[self.dname]['tvec']))
        self.data = ndata[self.dname]['Recs'][tind, ...]
        keys = ndata[self.dname].keys()
        shape = np.shape(self.data)
        if 'R' in keys:
            x = ndata[self.dname]['R']
            dx = (x[1] - x[0]) / 2
            self.xmin = x[0] - dx
            self.xmax = x[-1] + dx
        else:
            self.xmin = 0
            self.xmax = shape[1]
        if 'Z' in keys:
            z = ndata[self.dname]['Z']
            dz = (z[1] - z[0]) / 2
            self.ymin = z[0] - dz
            self.ymax = z[-1] + dz
        else:
            self.ymin = 0
            self.ymax = shape[0]
        if 'mag' in ndata:
            self.mdata = ndata['mag']['psiRz'][ndata['mtind']]
            self.mtvec = ndata['mag']['tflux'][ndata['mtind']]
            self.magchb.state(['!disabled'])
            self.magvar.set(1)
        else:
            self.magvar.set(0)
            self.magchb.state(['disabled'])
        try:
            self.bnd = np.loadtxt('tomotok/output/graphs/border.txt')
        except IOError:
            self.bndvar.set(0)
            self.bndchb.state(['disabled'])
        self.vmin = np.min(self.data)
        self.vmax = np.max(self.data)
        self.set_vals(vals=ndata[self.dname]['tvec'][tind])

    def get_mval(self, val):
        """
        Returns index for ploting magnetic data from index of reconstrution.
        Should not be called if magnetics data are not loaded.
        
        Parameters
        ----------
        val : int
            Index of reconstrition.
        """
        try:
            out = np.argmin(np.abs(self.mtvec - self.vals[val]))
        except:
            print('get_mval failure')
            out = 0
        return out

    def save(self, *args):
        fname = 'Graphs'
        if not os.path.exists(fname):
            os.makedirs(fname)
        val = self.scale.get()
        tm = np.round(1000 * self.vals[val])
        self.fig.savefig(os.path.join(fname, '{}at{:.0f}ms_{}.png'.format(
            self.shot, tm, self.dname)))


class GraphRgb(GraphWindow):
    """
    Based on GraphWindow class from widgets.py. Usefull for comparing
    tomography results from different diagnostics simultaneously.

    See documentation of Graf_window for information about other keyword arguments.
    
    Parameters
    ----------
    parent : TKInter class parent
        Default is None.
    data : dict
        Contains data, default is empty dict.
    
    Keywords
    --------
    title : str
        Name of graph window. Is used for saving. Default is 'RGB plot'
    """

    def __init__(self, parent=None, data=None, **kw):
        GraphWindow.__init__(self, parent, **kw)
        if data is None:
            data = {}
        # self.col = {'bol': 'r', 'ntr': 'g', 'sxr': 'b'}
        self.col = {'bol': 'c', 'ntr': 'm', 'sxr': 'y'}
        self.magvar = tk.BooleanVar(value=1)
        self.magchb = ttk.Checkbutton(self.bbox, text='Magnetics',
                                      variable=self.magvar, onvalue=True,
                                      command=self.update_graph)
        self.magchb.grid(row=self.bb_row, sticky='W')
        self.bb_row_raise()
        if 'title' in kw:
            self.title(kw['title'])
        else:
            self.title('RGB plot')
        #            self.title('CMY plot')
        self.update_data(data, vals=self.vals)

    def update_data(self, ndata, **kw):
        """
        Loads new data.
        
        Parameters
        ----------
        ndata : dict
            New data to be loaded to Graph window.
        """
        #        self.data = {}
        self.tvecs = {}
        self.diags = []
        try:
            self.shot = ndata['shot']
        except KeyError:
            self.shot = 0
        keys = ndata.keys()
        if 'tind' not in keys:
            ndata['tind'] = {}
        if 'bol' in keys:
            if 'bol' in ndata['tind']:
                tind = ndata['tind']['bol']
            else:
                tind = range(len(ndata['bol']['tvec']))
            bol = ndata['bol']['Recs'][tind, ...]
            nbol = bol / np.max(bol)
            nbol = np.where(nbol < 0, 0, nbol)
            self.tvecs['bol'] = ndata['bol']['tvec'][tind]
            self.diags.append('bol')
        else:
            bol = None
            nbol = None
            self.tvecs['bol'] = np.zeros(1)
        if 'ntr' in keys:
            if 'ntr' in ndata['tind']:
                tind = ndata['tind']['ntr']
            else:
                tind = range(len(ndata['sxr']['tvec']))
            ntr = ndata['ntr']['Recs'][tind, ...]
            nntr = ntr / np.max(ntr)
            nntr = np.where(nntr < 0, 0, nntr)
            self.tvecs['ntr'] = ndata['ntr']['tvec'][tind]
            self.diags.append('ntr')
        else:
            ntr = None
            nntr = None
            self.tvecs['ntr'] = np.zeros(1)
        if 'sxr' in keys:
            if 'sxr' in ndata['tind']:
                tind = ndata['tind']['sxr']
            else:
                tind = range(len(ndata['sxr']['tvec']))
            sxr = ndata['sxr']['Recs'][tind, ...]
            nsxr = sxr / np.max(sxr)
            nsxr = np.where(nsxr < 0, 0, nsxr)
            self.tvecs['sxr'] = ndata['sxr']['tvec'][tind]
            self.diags.append('sxr')
        else:
            sxr = None
            nsxr = None
            self.tvecs['sxr'] = np.zeros(1)
        if 'mag' in ndata:
            self.mdata = ndata['mag']['psiRz'][ndata['mtind']]
            self.mtvec = ndata['mag']['tflux'][ndata['mtind']]
            self.magvar.set(True)
            self.magchb.state(['!disabled'])
        else:
            self.magvar.set(False)
            self.magchb.state(['disabled'])
        if 'mag' in kw:
            if not kw['mag']:
                self.magvar.set(0)
                self.magchb.state(['disabled'])
        self.data = {'bol': nbol, 'ntr': nntr, 'sxr': nsxr}
        self.odata = {'bol': bol, 'ntr': ntr, 'sxr': sxr}
        self.key = self.diags[0]
        tvec = np.array(self.tvecs[self.key])
        keys = ndata[self.key]
        shape = np.shape(self.data[self.key])
        self.grids = [False, False]
        if 'R' in keys:
            x = ndata[self.key]['R']
            dx = (x[1] - x[0]) / 2
            self.xmin = x[0] - dx
            self.xmax = x[-1] + dx
            self.dx = 2 * dx
            self.grids[0] = True
        else:
            self.xmin = 0
            self.xmax = shape[1]
            self.dx = 1
        if 'Z' in keys:
            z = ndata[self.key]['Z']
            dz = (z[1] - z[0]) / 2
            self.ymin = z[0] - dz
            self.ymax = z[-1] + dz
            self.dy = 2 * dz
            self.grids[1] = True
        else:
            self.ymin = 0
            self.ymax = shape[0]
            self.dy = 1
        self.extent = [self.xmin, self.xmax, self.ymin, self.ymax]
        self.shape = shape
        if len(self.diags) < 3:
            shape = np.array(shape)
            shape[0] = 1
            if not 'bol' in self.diags:
                nbol = np.zeros(shape)
                bol = np.zeros(shape)
                self.tvecs['bol'] = np.zeros(1)
            if not 'ntr' in self.diags:
                nntr = np.zeros(shape)
                ntr = np.zeros(shape)
                self.tvecs['ntr'] = np.zeros(1)
            if not 'sxr' in self.diags:
                nsxr = np.zeros(shape)
                sxr = np.zeros(shape)
                self.tvecs['sxr'] = np.zeros(1)
            self.data = {'bol': nbol, 'ntr': nntr, 'sxr': nsxr}
            self.odata = {'bol': bol, 'ntr': ntr, 'sxr': sxr}
        self.set_vals(vals=tvec)

    #        self.plot(0)

    def get_mval(self, val):
        """
        Returns index for ploting magnetic data from index of reconstrution.
        Should not be called if magnetics data are not loaded.
        
        Parameters
        ----------
        val : int
            Index of reconstrition.
        """
        try:
            out = np.argmin(np.abs(self.mtvec - self.vals[val]))
        except:
            print('get_mval failure')
            out = 0
        return out

    def get_rgbval(self, val):
        """
        Gets indexes for given timeslices. Independent on provided diagnostics.
        
        Returns
        -------
        numpy.ndarray of ints
            Contains indexes of nearest tim vector values
            to reference diagnostic.
        """
        out = np.zeros(3, dtype=np.int)
        out[0] = np.argmin(np.abs(self.tvecs['bol'] - self.vals[val]))
        out[1] = np.argmin(np.abs(self.tvecs['ntr'] - self.vals[val]))
        out[2] = np.argmin(np.abs(self.tvecs['sxr'] - self.vals[val]))
        return out

    def plot(self, val, *args):
        inds = self.get_rgbval(val)
        b = self.data['bol'][inds[0]]
        n = self.data['ntr'][inds[1]]
        s = self.data['sxr'][inds[2]]
        rgb = np.stack((b, n, s), axis=2)
        cmy = np.ones_like(rgb) - rgb
        #        cmy = np.power(cmy, 2)
        self.ax.cla()
        for i in self.diags:
            self.ax.plot([], c=self.col[i], ls='', marker='s', label=i)
        self.ax.legend()
        self.ax.set_xlabel('R [m]')
        self.ax.set_ylabel('z [m]')
        self.ax.imshow(rgb, extent=[self.xmin, self.xmax, self.ymin, self.ymax], origin='lower')
        self.ax.imshow(cmy, extent=[self.xmin, self.xmax, self.ymin, self.ymax], origin='lower')
        if self.magvar.get():
            mval = self.get_mval(val)
            self.ax.contour(self.mdata[mval], [1.0],
                            # colors='w',
                            colors='k',
                            linewidths=1,
                            extent=[self.xmin, self.xmax, self.ymin, self.ymax])
        self.gcanvas.draw()

    def save(self):
        fname = 'Graphs'
        val = self.scale.get()
        if not os.path.exists(fname):
            os.makedirs(fname)
        self.fig.savefig(os.path.join(fname, '{} RGB plot at {:.0f}ms.png'.format(self.shot, self.vals[val] * 1000)))


class GraphPower(GraphWindow):
    """
    Based on GraphWindow class from widgets.py
    Vals are not time steps for this class but types of diagnotics to plot 
    power emitted or to show correlations of these.
    
    Parameters
    ----------
    parent : TKInter class parent, optional
        Default is None.
    data : dict, optional
        Contains data, default is empty dict.
    """

    # TODO correlation shifts?
    def __init__(self, parent=None, data=None, **kw):
        # self.vals = ['all', 'bol', 'ntr', 'sxr', 'preproc' ,'corel', 'shifts']
        GraphWindow.__init__(self, parent, **kw)
        if data is None:
            data = []
        self.title('Power graph')
        self.normvar = tk.BooleanVar(value=0)
        self.normchb = ttk.Checkbutton(self.bbox, text='Normalize',
                                       variable=self.normvar, onvalue=True,
                                       command=self.normalize)
        self.normchb.grid(row=self.bb_row, column=0)
        self.bb_row_raise()
        self.update_data(data, )
        self.plot(0)

    def plot(self, val, **kw):
        """
        Prepares data for matplotlib and draws them onto canvas
        
        Parameters
        ----------
        val : int
            number of slice in this case meaning type of analysis
        
        """
        if self.normvar.get():
            nrm = ' normalized'
        else:
            nrm = ''
        # col = {'bol': 'r', 'ntr': 'g', 'sxr': 'b'}
        col = {'bol': 'c', 'ntr': 'm', 'sxr': 'y'}
        ndiag = len(self.diags)
        ncor = len(self.cdata)
        shot = self.shot
        if val == 0:
            self.ax.cla()
            self.ax.set_title('#{} Total power{}'.format(shot, nrm))
            for i in range(ndiag):
                diag = self.diags[i]
                self.ax.plot(self.tvecs[i], self.data[i],
                             label=diag, color=col[diag])
            self.ax.legend()
            self.ax.set_xlabel('Time [s]')
            self.ax.set_ylabel('Power [{}]'.format(np.where(
                len(nrm) > 0, '-', 'Watt')))
            self.gcanvas.draw()
        elif 0 < val < ndiag:
            diag = self.vals[val]
            self.ax.cla()
            self.ax.set_title('#{} Total power of {}{}'.format(shot, diag, nrm))
            self.ax.set_xlabel('Time [s]')
            self.ax.set_ylabel('Power [{}]'.format(np.where(
                len(nrm) > 0, '-', 'Watt')))
            self.ax.plot(self.tvecs[val - 1], self.data[val - 1],
                         color=col[diag])
            self.gcanvas.draw()
        elif val == ndiag:
            self.ax.cla()
            self.ax.set_title('Preprocessed signals')
            self.ax.set_xlabel('Time [s]')
            for i in range(ndiag):
                diag = self.diags[i]
                self.ax.plot(self.tvecs[i], self.tmpdata[i],
                             label='Preproc ' + diag,
                             color=col[diag])
            self.ax.legend()
            self.gcanvas.draw()
        elif val == ndiag + 1:
            self.ax.cla()
            self.ax.set_title('Correlations{}'.format(nrm))
            self.ax.set_xlabel('Time shift [s]')
            lbls = ['corr(bolo,ntr)', 'corr(sxr,bolo)', 'corr(ntr,sxr)']
            for i in range(ncor):
                self.ax.plot(self.ctvec, self.cdata[i],
                             label=lbls[i], color=col[i])
            self.ax.legend()
            self.gcanvas.draw()
        elif val == ndiag + 2:
            self.ax.cla()
            self.ax.set_title('Shifted signals{}'.format(nrm))
            self.ax.set_xlabel('Arbitrary Time [s]')
            for i in range(ndiag):
                self.ax.plot(self.stvec[i], self.data[i],
                             label='{} shifted {:0.2} s'.format(
                                 self.diags[i],
                                 abs(self.shifts[i])),
                             color=col[i])
            self.ax.legend()
            self.gcanvas.draw()
        return

    def update_data(self, ndata, **kwargs):
        """
        Updates class data
        
        Parameters
        ----------
        ndata : dict
            Dictionary containign new data under keys 'bol', 'ntr', 'sxr'.
        """
        data = ndata
        self.diags = []
        self.sdata = []
        self.cdata = []
        self.tvecs = []
        try:
            self.shot = data['shot']
        except KeyError:
            self.shot = 0
        keys = ndata.keys()
        if 'tind' not in keys:
            ndata['tind'] = {}
        tkeys = ndata['tind'].keys()
        if 'bol' in keys:
            if 'bol' in tkeys:
                tind = ndata['tind']['bol']
            else:
                tind = np.arange(len(ndata['bol']['tvec']))
            if 'TOPO' in data['bol']:
                ptot = 'TOPO'
            else:
                ptot = 'TOBU'
            b = data['bol'][ptot][tind]
            self.sdata.append(b)
            self.tvecs.append(data['bol']['tvec'][tind])
            self.diags.append('bol')
        if 'ntr' in keys:
            if 'ntr' in tkeys:
                tind = ndata['tind']['ntr']
            else:
                tind = np.arange(len(ndata['ntr']['tvec']))
            if 'TOPO' in data['ntr']:
                ptot = 'TOPO'
            else:
                ptot = 'TOBU'
            n = data['ntr'][ptot][tind]
            self.sdata.append(n)
            self.tvecs.append(data['ntr']['tvec'][tind])
            self.diags.append('ntr')
        if 'sxr' in keys:
            if 'sxr' in tkeys:
                tind = ndata['tind']['sxr']
            else:
                tind = np.arange(len(ndata['sxr']['tvec']))
            if 'TOPO' in data['sxr']:
                ptot = 'TOPO'
            else:
                ptot = 'TOBU'
            s = data['sxr'][ptot][tind]
            self.sdata.append(s)
            self.tvecs.append(data['sxr']['tvec'][tind])
            self.diags.append('sxr')
        self.normd = None
        self.data = self.sdata
        self.set_vals(vals=['all'] + self.diags)
        self.normalize()

    def normalize(self, *args, **kw):
        """
        Switches data to normalized or standard, depending on norvar state.
        Normalizes data and if no normalized data are available.
        """
        if self.normd is None:
            self.normd = [i / np.max(i) for i in self.data]
        if self.normvar.get():
            self.data = self.normd
        else:
            self.data = self.sdata
        self.update_graph()

    def save(self):
        fname = 'Graphs'
        if not os.path.exists(fname):
            os.makedirs(fname)
        self.fig.savefig(os.path.join(fname, '{} {}.png'.format(self.shot, self.ax.get_title())))
        # self.fig.savefig(os.path.join(fname, '{} {}.eps'.format(self.shot, self.ax.get_title())))
